"""
Bayesian calibration of hierarchical fission gas release models
===============================================================
"""

# %%
# Introduction
# ------------
#
# Models such as the ones described in :ref:`use-case-fission-gas` include empirical constants that need to be calibrated so that predictions agree well with measurements.
# During this process, the distributions of model parameters are derived,
# which can be used in subsequent analyses to generate accurate forecasts with corresponding uncertainties.
# However, model inadequacy can disrupt calibration,
# leading to derived uncertainties that fail to cover prediction-measurement discrepancies.
# A Bayesian way to account for this is to assume that the model parameters vary between different sets of experimental conditions, i.e.
# to adopt a hierarchical model and calibrate distribution parameters (means and standard deviations).
#
# Using the notations defined in :ref:`use-case-fission-gas`,
# the unobserved model inputs :math:`x_{\mathrm{diff}, i}, i=1...\sampleSize_{\mathrm{exp}}`
# (resp. :math:`x_{\mathrm{crack}, i}, i=1...\sampleSize_{\mathrm{exp}}`)
# are i.i.d. random variables which follow a normal distribution with
# mean parameter :math:`\mu_{\mathrm{diff}}` (resp. :math:`\mu_{\mathrm{crack}}`)
# and standard deviation parameter :math:`\sigma_{\mathrm{diff}}` (resp. :math:`\sigma_{\mathrm{crack}}`).
#
# The network plot from the page :ref:`use-case-fission-gas` can thus be updated:
#
# .. figure:: ../../_static/fission_gas_network_calibration.png
#     :align: center
#     :alt: Fission gas release calibration
#     :width: 50%
#
# In the network above, full arrows represent deterministic relationships and dashed arrows probabilistic relationships.
# More precisely, the conditional distribution of the node at the end of two dashed arrows when (only) the starting nodes are known
# is a normal distribution with parameters equal to these starting nodes.
# Note that due to constraints on the input domains of the :math:`\model_i` models, for every :math:`1 \leq i \leq \sampleSize_{\mathrm{exp}}`,
# the distributions of :math:`x_{\mathrm{diff}, i}` and :math:`x_{\mathrm{crack}, i}` are truncated to the input domain boundaries.
#
# Such a hierarchical approach was used in [robertson2024]_, showing how a hierarchical probabilistic model can be sampled using a Metropolis-Hastings-within-Gibbs sampler.
# The authors calibrated the models against measurements from the International Fuel Performance Experiments (IFPE) database.
# This example follows the procedure described in their paper.
# Please note however that we are using simplified models, so the results of this page
# should not be expected to reproduce those of the paper.

import openturns as ot
import openturns.viewer as otv
import numpy as np
import matplotlib.pyplot as plt

# %%
# Load the models

from openturns.usecases import fission_gas

fgr = fission_gas.FissionGasRelease()
desc = fgr.getInputDescription()  # description of the model inputs (diff, crack)
ndim = len(desc)  # dimension of the model inputs: 2
nexp = fgr.measurement_values.getSize()  # number of sets of experimental conditions
models = fgr.models  # the nexp models

# %%
# Each experiment :math:`i` produced one measurement value,
# which is used to define the likelihood of the associated model :math:`\model_i`
# and latent variables :math:`x_{i, \mathrm{diff}}` and :math:`x_{i, \mathrm{crack}}`.

likelihoods = [
    ot.Normal(v, fgr.measurement_uncertainty(v)) for v in fgr.measurement_values
]

# %%
# Partially conjugate posterior
# -----------------------------
#
# The goal of this study is to calibrate the parameters :math:`\mu_{\mathrm{diff}}`, :math:`\sigma_{\mathrm{diff}}`,
# :math:`\mu_{\mathrm{crack}}` and :math:`\sigma_{\mathrm{crack}}`.
# To perform Bayesian calibration, we set for :math:`\mu_{\mathrm{diff}}` and :math:`\mu_{\mathrm{crack}}` a uniform prior distribution
# and for :math:`\sigma_{\mathrm{diff}}` and :math:`\sigma_{\mathrm{crack}}`
# the limit of a truncated inverse gamma distribution with parameters :math:`(\lambda, k)` when :math:`\lambda \to \infty` and :math:`k \to 0`.
# The parameters of the prior distributions are defined later.
#
# This choice of prior distributions means that the posterior is partially conjugate.
# For instance, the conditional posterior distribution of :math:`\mu_{\mathrm{diff}}`
# (resp. :math:`\mu_{\mathrm{crack}}`)
# is a truncated normal distribution with the following parameters
# (for :math:`\mu_{\mathrm{crack}}` simply replace :math:`\mathrm{diff}` with :math:`\mathrm{crack}` in what follows) :
#
# - The truncation parameters are the bounds of the prior uniform distribution.
# - The mean parameter is :math:`\frac{1}{\sampleSize_{\mathrm{exp}}} \sum_{i=1}^{\sampleSize_{\mathrm{exp}}} x_{\mathrm{diff}, i}`.
# - The standard deviation parameter is :math:`\sqrt{\frac{\sigma_{\mathrm{diff}}}{\sampleSize_{\mathrm{exp}}}}`.
#
# Let us prepare a random vector to sample the conditional posterior
# distributions of :math:`\mu_{\mathrm{diff}}` and :math:`\mu_{\mathrm{crack}}`.

mu_rv = ot.RandomVector(ot.TruncatedNormal())
mu_desc = [f"$\\mu$_{{{label}}}" for label in desc]

# %%
# The conditional posterior distribution of :math:`\sigma_{\mathrm{diff}}`
# (resp. :math:`\sigma_{\mathrm{crack}}`)
# is a truncated inverse gamma distribution with the following parameters
# (for :math:`\sigma_{\mathrm{crack}}` simply replace :math:`\mathrm{diff}` with :math:`\mathrm{crack}` in what follows) :
#
# - The truncation parameters are the truncation parameters of the prior distribution.
# - The :math:`\lambda` parameter is :math:`\frac{2}{\sum_{i=1}^{\sampleSize_{\mathrm{exp}}} \left(x_{\mathrm{diff}, i} - \mu_{\mathrm{diff}} \right)^2}`.
# - The :math:`k` parameter is :math:`\sqrt{\frac{\sampleSize_{\mathrm{exp}}}{2}}`.
#
# Let us prepare a random vector to sample the conditional posterior
# distribution of :math:`\sigma_{\mathrm{diff}}^2` and :math:`\sigma_{\mathrm{crack}}^2`.

sigma_square_rv = ot.RandomVector(ot.TruncatedDistribution(ot.InverseGamma(), 0.0, 1.0))
sigma_square_desc = [f"$\\sigma$_{{{label}}}^2" for label in desc]


# %%
# We define 3 function templates which produce:
#
# - the parameters of the (truncated normal) conditional posterior distributions of the :math:`\mu` parameters
# - the parameters of the (truncated inverse gamma) conditional posterior distributions of the :math:`\sigma` parameters
# - the conditional posterior log-PDF of the latent variables.


class PosteriorParametersMu(ot.OpenTURNSPythonFunction):
    """Outputs the parameters of the conditional posterior distribution of one
    of the mu parameters.
    This conditional posterior distribution is a TruncatedNormal distribution.

    Parameters
    ----------
    state : list of float
        Current state of the Markov chain.
        The posterior distribution is conditional to those values.

    Returns
    -------
    parameters : list of float
        Parameters of the conditional posterior distribution.
        In order: mean, standard deviation, lower bound, upper bound.
    """

    def __init__(self, dim=0, lb=-100, ub=100):
        self._dim = dim

        # State description: mu values, then sigma values, then for each experiment x values
        state_length = (1 + 1 + nexp) * ndim
        super().__init__(state_length, 4)
        self._xindices = range(state_length)[2 * ndim :][dim::ndim]

        # Get lower and upper bound
        self._lb = lb
        self._ub = ub

    def _exec(self, state):
        # posterior mean of mu = empirical mean of the x values
        post_mean = np.mean([state[i] for i in self._xindices])

        # posterior std of mu = prior sigma / sqrt(nexp)
        post_std = np.sqrt(state[ndim + self._dim] / nexp)

        # Hyperparameters of a truncated normal
        return [post_mean, post_std, self._lb, self._ub]


class PosteriorParametersSigmaSquare(ot.OpenTURNSPythonFunction):
    """Outputs the parameters of the conditional posterior distribution of one
    of the sigma parameters.
    This conditional posterior distribution is a Truncated InverseGamma distribution.

    Parameters
    ----------
    state : list of float
        Current state of the Markov chain.
        The posterior distribution is conditional to those values.

    Returns
    -------
    parameters : list of float
        Parameters of the conditional posterior distribution.
        In order: k (shape), lambda, lower bound, upper bound.
    """

    def __init__(self, dim=0, lb=1e-4, ub=100):
        self._dim = dim

        # State description: mu values, then sigma values, then for each experiment x values
        state_length = (1 + 1 + nexp) * ndim
        super().__init__(state_length, 4)
        self._xindices = range(state_length)[2 * ndim :][dim::ndim]

        # Get lower and upper bound
        self._lb = lb
        self._ub = ub

    def _exec(self, state):
        mu = state[self._dim]

        # Get squares of centered xvalues from the state
        squares = [(state[i] - mu) ** 2 for i in self._xindices]

        post_lambda = 2.0 / np.sum(squares)  # rate lambda =  1 / beta
        post_k = nexp / 2.0  # shape

        return [post_k, post_lambda, self._lb, self._ub]


class PosteriorLogDensityX(ot.OpenTURNSPythonFunction):
    """Outputs the conditional posterior density (up to an additive constant)
    of the 2D latent variable x_i = (x_{i, diff}, x_{i, x_{i, crack})
    corresponding to one experiment i.

    Parameters
    ----------
    state : list of float
        Current state of the Markov chain.
        The posterior distribution is conditional to those values.

    Returns
    -------
    log_density : list of one float
        PLog-density (up to an additive constant) of the conditional posterior.
    """

    def __init__(self, exp):
        # State description: mu values, then sigma values, then for each experiment x values
        state_length = (1 + 1 + nexp) * ndim
        super().__init__(state_length, 1)
        self._xindices = range(state_length)[2 * ndim :][exp * ndim : (exp + 1) * ndim]

        # Setup experiment number and associated model and likelihood
        self._exp = exp
        self._like = likelihoods[exp]
        self._model = models[exp]

    def _exec(self, state):
        # Get the x indices of the experiment
        x = np.array([state[i] for i in self._xindices])

        # Get mu and sigma in order to normalize x
        mu = np.array([state[i] for i in range(ndim)])
        sig = np.sqrt([state[i] for i in range(ndim, 2 * ndim)])
        normalized = (x - mu) / sig

        # Compute the log-prior density
        logprior = np.sum([ot.DistFunc.logdNormal(normalized[i]) for i in range(ndim)])

        # Use the model to predict the experiment and compute the log-likelihood
        pred = self._model(x)
        loglikelihood = self._like.computeLogPDF(pred)

        # Return the log-posterior, i.e. the sum of the log-prior and the log-likelihood
        return [logprior + loglikelihood]


# %%
# Metropolis-within-Gibbs algorithm
# ---------------------------------
#
# Lower and upper bounds for :math:`\mu_{\mathrm{diff}}, \mu_{\mathrm{crack}}`
lbs = [0.1, 1e-4]
ubs = [40.0, 1.0]

# %%
# Lower and upper bounds for :math:`\sigma_{\mathrm{diff}}^2, \sigma_{\mathrm{crack}}^2`
lbs_sigma_square = np.array([0.1, 0.1]) ** 2
ubs_sigma_square = np.array([40, 10]) ** 2

# %%
# Initial state
initial_mus = [10.0, 0.3]
initial_sigma_squares = [20.0**2, 0.5**2]
initial_x = np.repeat([[19.0, 0.4]], repeats=nexp, axis=0).flatten().tolist()
initial_state = initial_mus + initial_sigma_squares + initial_x

# %%
# Support of the prior (and thus posterior) distribution
support = ot.Interval(
    lbs + lbs_sigma_square.tolist() + nexp * lbs,
    ubs + ubs_sigma_square.tolist() + nexp * ubs,
)

# %%
# Create the list of all samplers in the Gibbs algorithm as outlined in the chart below,
# where direct samplers are represented in green and random walk Metropolis-Hastings samplers in blue.
#
# .. figure:: ../../_static/fission_gas_mh.png
#     :align: center
#     :alt: Metropolis-Hastings-within-Gibbs description
#     :width: 100%
#
# We start with the samplers of :math:`\mu_{\mathrm{diff}}, \mu_{\mathrm{crack}}`.
# We are able to directly sample these conditional distributions,
# hence we use the :class:`~openturns.RandomVectorMetropolisHastings` class.

samplers = [
    ot.RandomVectorMetropolisHastings(
        mu_rv,
        initial_state,
        [i],
        ot.Function(PosteriorParametersMu(dim=i, lb=lbs[i], ub=ubs[i])),
    )
    for i in range(ndim)
]

# %%
# We continue with the samplers of :math:`\sigma_{\mathrm{diff}}^2, \sigma_{\mathrm{crack}}^2`.
# We are also able to directly sample these conditional distributions.

samplers += [
    ot.RandomVectorMetropolisHastings(
        sigma_square_rv,
        initial_state,
        [ndim + i],
        ot.Function(
            PosteriorParametersSigmaSquare(
                dim=i, lb=lbs_sigma_square[i], ub=ubs_sigma_square[i]
            )
        ),
    )
    for i in range(ndim)
]

# %%
# We finish with the samplers of the :math:`\vect{x}_i`, with :math:`1 \leq i \leq n_{exp}`.
# Each of these samplers outputs points in a :math:`\sampleSize_{\mathrm{exp}}`-dimensional space.
# We are not able to directly sample these conditional posterior distributions,
# so we resort to random walk Metropolis-Hastings.

for exp in range(nexp):
    base_index = 2 * ndim + ndim * exp

    samplers += [
        ot.RandomWalkMetropolisHastings(
            ot.Function(PosteriorLogDensityX(exp=exp)),
            support,
            initial_state,
            ot.Normal([0.0] * 2, [6.0, 0.15]),
            [base_index + i for i in range(ndim)],
        )
    ]

# %%
# The Gibbs algorithm combines all these samplers.

sampler = ot.Gibbs(samplers)
x_desc = []
for i in range(1, nexp + 1):
    x_desc += [f"x_{{{label}, {i}}}" for label in desc]
sampler.setDescription(mu_desc + sigma_square_desc + x_desc)

# %%
# Run this Metropolis-within-Gibbs algorithm and check the acceptance rates
# for the Random walk Metropolis-Hastings samplers.

samples = sampler.getSample(2000)
acceptance = [
    sampler.getMetropolisHastingsCollection()[i].getAcceptanceRate()
    for i in range(6, len(samplers))
]

adaptation_factor = [
    sampler.getMetropolisHastingsCollection()[i]
    .getImplementation()
    .getAdaptationFactor()
    for i in range(6, len(samplers))
]

print("Minimum acceptance rate = ", np.min(acceptance))
print("Maximum acceptance rate for random walk MH = ", np.max(acceptance[2 * ndim :]))

# %%
# Plot the posterior distribution
# -------------------------------
#
# Please note that the following plots rely on the MCMC sample.
# Although this is not done in the present example,
# diagnostics should be run on the MCMC sample to assess the convergence of the Markov chain.
#
# We only represent the :math:`\mu` and :math:`\sigma` parameters.

reduced_samples = samples[:, 0:4]

# %%
# It is possible to quickly draw pair plots.
# Here we tweak the rendering a little.

pair_plots = ot.VisualTest.DrawPairs(reduced_samples)

for i in range(pair_plots.getNbRows()):
    for j in range(pair_plots.getNbColumns()):
        graph = pair_plots.getGraph(i, j)
        graph.setXTitle(pair_plots.getGraph(pair_plots.getNbRows() - 1, j).getXTitle())
        graph.setYTitle(pair_plots.getGraph(i, 0).getYTitle())
        pair_plots.setGraph(i, j, graph)

_ = otv.View(pair_plots)

# %%
# Create an enhanced pair plots grid with histograms of the marginals on the diagonal.

full_grid = ot.GridLayout(pair_plots.getNbRows() + 1, pair_plots.getNbColumns() + 1)

for i in range(full_grid.getNbRows()):
    hist = ot.HistogramFactory().build(reduced_samples.getMarginal(i))
    pdf = hist.drawPDF()
    pdf.setLegends([""])
    pdf.setTitle("")
    full_grid.setGraph(i, i, pdf)

for i in range(pair_plots.getNbRows()):
    for j in range(pair_plots.getNbColumns()):
        if len(pair_plots.getGraph(i, j).getDrawables()) > 0:
            full_grid.setGraph(i + 1, j, pair_plots.getGraph(i, j))

_ = otv.View(full_grid)


# %%
# Finally superimpose contour plots of the KDE-estimated 2D marginal PDFs on the pairplots.

ot.ResourceMap.SetAsBool("Contour-DefaultIsFilled", True)

# sphinx_gallery_thumbnail_number = 3
for i in range(1, full_grid.getNbRows()):
    for j in range(i):
        graph = full_grid.getGraph(i, j)
        cloud = graph.getDrawable(0).getImplementation()
        cloud.setPointStyle(".")
        data = cloud.getData()
        dist = ot.KernelSmoothing().build(data)
        contour = dist.drawPDF().getDrawable(0).getImplementation()
        contour.setLevels(np.linspace(0.0, contour.getLevels()[-1], 10))
        graph.setDrawables([contour, cloud])
        graph.setBoundingBox(contour.getBoundingBox())
        full_grid.setGraph(i, j, graph)

_ = otv.View(full_grid, scatter_kw={"alpha": 0.1})


# %%
# Post-calibration prediction
# ---------------------------
#
# Retrieve the :math:`\mu` and :math:`\sigma^2` columns in the sample.

mu = samples.getMarginal([f"$\\mu$_{{{label}}}" for label in desc])
sigma_square = samples.getMarginal([f"$\\sigma$_{{{label}}}^2" for label in desc])


# %%
# Build the joint distribution of the latent variables :math:`x_{\mathrm{diff}}, x_{\mathrm{crack}}`
# obtained when :math:`\mu_{\mathrm{diff}}`, :math:`\sigma_{\mathrm{diff}}`,
# :math:`\mu_{\mathrm{crack}}` and :math:`\sigma_{\mathrm{crack}}`
# follow their joint posterior distribution.
# It is estimated as a mixture of truncated :math:`\sampleSize_{\mathrm{exp}}`-dimensional normal distributions
# corresponding to the posterior samples of the :math:`\mu_{\mathrm{diff}}`, :math:`\mu_{\mathrm{crack}}`,
# :math:`\sigma_{\mathrm{diff}}` and :math:`\sigma_{\mathrm{crack}}` parameters.

truncation_interval = ot.Interval(lbs, ubs)
normal_collection = [
    ot.TruncatedDistribution(ot.Normal(mean, np.sqrt(var)), truncation_interval)
    for (mean, var) in zip(mu, sigma_square)
]
normal_mixture = ot.Mixture(normal_collection)
normal_mixture.setDescription(desc)

# %%
# Build a collection of random vectors such that the distribution
# of each is the push-forward of the marginal distribution of :math:`(x_{\mathrm{diff}}, x_{\mathrm{crack}})`
# defined above through one of the :math:`\sampleSize_{\mathrm{exp}}` models.

rv_normal_mixture = ot.RandomVector(normal_mixture)
rv_models = [ot.CompositeRandomVector(model, rv_normal_mixture) for model in models]

# %%
# Get a Monte-Carlo estimate of the median, 0.05 quantile and 0.95 quantile
# of these push-forward distributions.

predictions = [rv.getSample(200) for rv in rv_models]
prediction_medians = [sam.computeMedian()[0] for sam in predictions]
prediction_lb = [sam.computeQuantile(0.05)[0] for sam in predictions]
prediction_ub = [sam.computeQuantile(0.95)[0] for sam in predictions]

# %%
# These push-forward distributions are the distributions
# of the model predictions when :math:`\mu_{\mathrm{diff}}`, :math:`\mu_{\mathrm{crack}}`,
# :math:`\sigma_{\mathrm{diff}}` and :math:`\sigma_{\mathrm{crack}}` follow
# their joint posterior distribution.
# They can be compared to the actual measurements to represent predictive accuracy.

yerr = np.abs(np.column_stack([prediction_lb, prediction_ub]).T - prediction_medians)
plt.errorbar(fgr.measurement_values, prediction_medians, yerr, fmt="o")
plt.xscale("log")

bisector = np.linspace(0, 0.5)
plt.plot(bisector, bisector, "--", color="black")

plt.xlabel("Measurements")
plt.ylabel("Prediction ranges induced by the posterior")

plt.show()

# %%
# For the sake of comparison, we now consider the distributions
# of the model predictions when :math:`\mu_{\mathrm{diff}}`, :math:`\mu_{\mathrm{crack}}`,
# :math:`\sigma_{\mathrm{diff}}` and :math:`\sigma_{\mathrm{crack}}` follow
# their joint prior distribution.
# Because the actual prior distribution of :math:`\sigma_{\mathrm{diff}}` and :math:`\sigma_{\mathrm{crack}}`
# cannot be represented, we approximate them by choosing a very large :math:`\lambda` parameter
# and a very small :math:`k` parameter.

prior = ot.JointDistribution(
    [
        ot.Uniform(lbs[0], ubs[0]),
        ot.Uniform(lbs[1], ubs[1]),
        ot.TruncatedDistribution(
            ot.InverseGamma(0.01, 1e7),
            lbs_sigma_square[0],
            float(ubs_sigma_square[0]),
        ),
        ot.TruncatedDistribution(
            ot.InverseGamma(0.01, 1e7),
            lbs_sigma_square[1],
            float(ubs_sigma_square[1]),
        ),
    ]
)
prior_sample = prior.getSample(2000)

# As before, build a mixture of truncated normal distributions from the sample.
normal_collection_prior = [
    ot.TruncatedDistribution(ot.Normal(par[0:2], np.sqrt(par[2:])), truncation_interval)
    for par in prior_sample
]
normal_mixture_prior = ot.Mixture(normal_collection_prior)
normal_mixture_prior.setDescription(desc)
rv_normal_mixture_prior = ot.RandomVector(normal_mixture_prior)

# Build random vectors sampling from the predictive distributions.
rv_models_prior = [
    ot.CompositeRandomVector(model, rv_normal_mixture_prior) for model in models
]

predictions_prior = [rv.getSample(100) for rv in rv_models_prior]
prediction_medians_prior = [sam.computeMedian()[0] for sam in predictions_prior]
prediction_lb_prior = [sam.computeQuantile(0.05)[0] for sam in predictions_prior]
prediction_ub_prior = [sam.computeQuantile(0.95)[0] for sam in predictions_prior]

# Produce the graph comparing predictive distributions with measurements
yerr_prior = np.abs(
    np.column_stack([prediction_lb_prior, prediction_ub_prior]).T
    - prediction_medians_prior
)
plt.errorbar(
    np.array(fgr.measurement_values), prediction_medians_prior, yerr_prior, fmt="o"
)
plt.xscale("log")
plt.plot(bisector, bisector, "--", color="black")
plt.xlabel("Measurements")
plt.ylabel("Prediction ranges induced by the prior")
plt.show()
# %%
# To facilitate the comparison, we plot the median value of the predictive distributions only.

plt.scatter(fgr.measurement_values, prediction_medians)
plt.scatter(fgr.measurement_values, prediction_medians_prior)
plt.xscale("log")
plt.plot(bisector, bisector, "--", c="black")
plt.xlabel("Measurements")
plt.ylabel("Prediction medians")
_ = plt.legend(["Posterior", "Prior"])

# %%
# Display all figures
otv.View.ShowAll()



import openturns as ot
from matplotlib import pyplot as plt
import openturns.viewer as otv

R = ot.CorrelationMatrix(2)
for i in range(1, 2):
    R[i, i - 1] = -0.25

distribution = ot.Normal([-1.5, 0.5], [4.0, 1.0], R)

sample = distribution.getSample(200)

# Create an empty graph
graph = ot.Graph("Cloud example", "x1", "x2")

# Create the cloud
cloud = ot.Cloud(sample, "blue", "fsquare", "")

# Then, draw it
graph.add(cloud)

fig = plt.figure(figsize=(4, 4))
axis = fig.add_subplot(111)
axis.set_xlim(auto=True)
otv.View(graph, figure=fig, axes=[axis], add_legend=False)



#!/usr/bin/env python
# An optimization problem with mixed discrete input variables.
#
# min f(x1,x2,x3)
# where
# x1 is discrete in {0,1,2,3,4}
# x2 and x3 are real in [0,1]
#
# Mixed parameter space with one discrete parameter \eqn{x_1 \in \{0, 1, 2, 3, 4\}}
# and two numerical parameters \eqn{x_1, x_2 \in [0, 1]}.
# The function is defined as follows:
# f(\mathbf{x}) = \\
# \sin(2\pi x_3 - \pi) + 7 \sin^2(2 \pi x_2 - \pi) \, if \, x_1 = 0 \\
# \sin(2\pi x_3 - \pi) + 7 \sin^2(2 \pi x_2 - \pi) + 12 \sin(2 \pi x_3 - \pi) \, if \, x_1 = 1 \\
# \sin(2\pi x_3 - \pi) + 7 \sin^2(2 \pi x_2 - \pi) + 0.5 \sin(2 \pi x_3 - \pi) \, if \, x_1 = 2 \\
# \sin(2\pi x_3 - \pi) + 7 \sin^2(2 \pi x_2 - \pi) + 8.0 \sin(2 \pi x_3 - \pi) \, if \, x_1 = 3 \\
# \sin(2\pi x_3 - \pi) + 7 \sin^2(2 \pi x_2 - \pi) + 3.5 \sin(2 \pi x_3 - \pi) \, if \, x_1 = 4.
#
# Reference
# Ported from R to Python, based on
# https://rdrr.io/cran/smoof/src/R/sof.swiler2014.R

import math
import openturns as ot
import openturns.testing as ott


def swiler2014(x):
    x1, x2, x3 = x
    # minimum  = f([1, 0 , 0.25])=-13

    a = math.sin(2 * math.pi * x3 - math.pi)
    b = 7 * math.sin(2 * math.pi * x2 - math.pi) ** 2
    facs = [0, 12.0, 0.5, 8.0, 3.5]  # A set of numeric values
    fac = facs[int(x1)]  # Convert to int, then use the value in the list
    result = a + b + fac * a
    # print("x=",[x1,x2,x3],"f(x)=",result)
    return [result]


# Create the objective function
objectiveFun = ot.PythonFunction(3, 1, swiler2014)
x = [1, 0, 0.25]
print("Evaluate f at x=", x)
print(f"f(x)={objectiveFun(x)}")

# Define OptimizationProblem
problem = ot.OptimizationProblem(objectiveFun)
bounds = ot.Interval([0.0] * 3, [1.0, 1.0, 4.0])
varTypes = [
    ot.OptimizationProblemImplementation.INTEGER,
    ot.OptimizationProblemImplementation.CONTINUOUS,
    ot.OptimizationProblemImplementation.CONTINUOUS,
]
problem.setBounds(bounds)
problem.setVariablesType(varTypes)
problem.setMinimization(True)

# Define OptimizationAlgorithm
x0 = [0.0] * 3
algo = ot.Bonmin(problem)
algo.setStartingPoint(x0)
algo.setMaximumCallsNumber(10000)
algo.setMaximumIterationNumber(1000)


for name in ot.Bonmin.GetAlgorithmNames():
    print(f"-- {name} algorithm...")
    algo.setAlgorithmName(name)
    algo.run()

    # Retrieve result
    result = algo.getResult()
    x_star = result.getOptimalPoint()
    print(f"x*={x_star}")
    y_star = result.getOptimalValue()
    neval = result.getCallsNumber()
    print(f"f(x*)={y_star} neval={neval}")
    ott.assert_almost_equal(x_star, [1.0, 0.0, 0.25], 1, 5e-4)



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()


distribution = ot.Normal(4)
distribution.setMu([5.0] * 4)
size = 16
experiment = ot.LowDiscrepancyExperiment(ot.HaltonSequence(), distribution, size)
print("experiment = ", experiment)
# Test sampling with weights
sample, weights = experiment.generateWithWeights()
print("sample  = ", sample)
print("weights = ", weights)
# Test sampling with reinitialization each time the distribution is
# set (default behaviour)
# sample 2 != sample
print("sample2 = ", experiment.generate())
experiment.setDistribution(distribution)
# sample 3 == sample
print("sample3 = ", experiment.generate())

# Test sampling without reinitialization excepted when distribution
# dimension changes
experiment = ot.LowDiscrepancyExperiment(ot.HaltonSequence(), distribution, size, False)
print("sample  = ", experiment.generate())
# sample 2 != sample
print("sample2 = ", experiment.generate())
experiment.setDistribution(distribution)
# sample 3 != sample && sample 3 != sample 2
print("sample3 = ", experiment.generate())
# Test dimension change
experiment.setDistribution(ot.Normal())
print("sample = ", experiment.generate())

# Test constructor with no distribution and dimension>1
experiment = ot.LowDiscrepancyExperiment(ot.HaltonSequence(2), size)
print("sample = ", experiment.generate())
# Test with dependent marginals
R = ot.CorrelationMatrix(4)
for i in range(1, 4):
    R[i - 1, i] = 0.5
distribution.setR(R)
experiment = ot.LowDiscrepancyExperiment(ot.HaltonSequence(), distribution, size, False)
print("sample = ", experiment.generate())



import openturns as ot
from matplotlib import pyplot as plt
import openturns.viewer as otv

conditioningDist = ot.Normal(0.0, 1.0)
link = ot.SymbolicFunction(["y"], ["y", "0.1+y^2"])
conditionedDist = ot.Normal()
distribution = ot.JointByConditioningDistribution(
    conditionedDist, conditioningDist, link
)
distribution.setDescription(["$Y$", "$X$"])

pdf_graph = distribution.drawPDF([-1.0] * 2, [1.0] * 2)
pdf_graph.setTitle(str(distribution))
fig = plt.figure(figsize=(10, 5))
pdf_axis = fig.add_subplot(111)
otv.View(pdf_graph, figure=fig, axes=[pdf_axis], add_legend=False)



"""
Gaussian Process Regression vs KrigingAlgorithm
================================================
"""

# %%
# The goal of this example is to highlight the main changes between the old API involving `KrigingAlgorithm` and the new one.
#
# It assumes a basic knowledge of Gaussian Process Regression.
# For that purpose, we create a Gaussian Process Regression surrogate model for a function which has scalar real inputs and outputs.
# We select a very simple example.

# %%
# Introduction
# ------------
#
# We consider the sine function:
#
# .. math::
#    y = x \sin(x)
#
#
# for any :math:`x\in[0,12]`.
#
# We want to create a surrogate of this function. This is why we create a sample of :math:`n` observations of the function:
#
# .. math::
#    y_i=x_i \sin(x_i)
#
# We are going to consider a Gaussian Process Regression with:
#
# * a constant trend,
# * a Matern covariance model.

# %%
import openturns as ot
import openturns.viewer as otv


# %%
# First let us introduce some useful function.
# In order to observe the function and the location of the points in the input design of experiments, we define `plot_1d_data`.


# %%
def plot_1d_data(x_data, y_data, type="Curve", legend=None, color=None, linestyle=None):
    """Plot the data (x_data,y_data) as a Cloud/Curve"""
    if type == "Curve":
        graphF = ot.Curve(x_data, y_data)
    else:
        graphF = ot.Cloud(x_data, y_data)
    if legend is not None:
        graphF.setLegend(legend)
    if color is not None:
        graphF.setColor(color)
    if linestyle is not None:
        graphF.setLineStyle(linestyle)
    return graphF


def computeQuantileAlpha(alpha):
    """
    Compute bilateral confidence interval of level 1-alpha of a gaussian distribution.
    """
    bilateralCI = ot.Normal().computeBilateralConfidenceInterval(1 - alpha)
    return bilateralCI.getUpperBound()[0]


def computeBoundsConfidenceInterval(y_test_hat, quantileAlpha, conditionalSigma):
    """
    Compute the 1-alpha confidence interval bounds.
    """
    dataLower = [
        [y_test_hat[i, 0] - quantileAlpha * conditionalSigma[i, 0]]
        for i in range(n_test)
    ]
    dataUpper = [
        [y_test_hat[i, 0] + quantileAlpha * conditionalSigma[i, 0]]
        for i in range(n_test)
    ]
    dataLower = ot.Sample(dataLower)
    dataUpper = ot.Sample(dataUpper)
    return dataLower, dataUpper


# %%
g = ot.SymbolicFunction(["x"], ["x * sin(x)"])

# %%
xmin = 0.0
xmax = 12.0
n_train = 20
step = (xmax - 1 - xmin) / (n_train - 1.0)
x_train = ot.RegularGrid(xmin + 0.2, step, n_train).getVertices()
y_train = g(x_train)
n_train = x_train.getSize()

# %%
# In order to compare the function and its metamodel, we use a test (i.e. validation) design of experiments made of a regular grid of 100 points from 0 to 12.
# Then we convert this grid into a :class:`~openturns.Sample` and we compute the outputs of the function on this sample.

# %%
n_test = 100
step = (xmax - xmin) / (n_test - 1)
myRegularGrid = ot.RegularGrid(xmin, step, n_test)
x_test = myRegularGrid.getVertices()
y_test = g(x_test)

# %%
# We plot the true function (continuous dashed curve) and train data (cloud points) on the same figure.
graph = ot.Graph("Function of interest", "", "")
graph.add(
    plot_1d_data(x_test, y_test, legend="Exact", color="black", linestyle="dashed")
)
graph.add(
    plot_1d_data(x_train, y_train, type="Cloud", legend="Train points", color="red")
)
graph.setAxes(True)
graph.setXTitle("X")
graph.setYTitle("Y")
graph.setLegendPosition("upper right")
view = otv.View(graph)

# %%
# We use the :class:`~openturns.ConstantBasisFactory` class to define the trend and the :class:`~openturns.MaternModel` class to define the covariance model.
# This Matérn model is based on the regularity parameter :math:`\nu=3/2`.

# %%
dimension = 1
basis = ot.ConstantBasisFactory(dimension).build()
covarianceModel = ot.MaternModel([1.0] * dimension, 1.5)

# %%
# In the following, we use the `KrigingAlgorithm` class to fit the Gaussian Process Regression model (aka Kriging).

# %%
kriging_algo = ot.KrigingAlgorithm(x_train, y_train, covarianceModel, basis)
kriging_algo.run()
kriging_result = kriging_algo.getResult()
krigingMM = kriging_result.getMetaModel()

# %%
# We observe that the `scale` and `amplitude` hyper-parameters have been optimized by the `run` method.
# The default optimization method (used here) is the :class:`~openturns.TNC`
# With the new API, the :class:`~openturns.GaussianProcessFitter` class  is used to fit the
# gaussian process and :class:`~openturns.GaussianProcessRegression` to get the conditioned model.

# %%
fitter_algo = ot.GaussianProcessFitter(x_train, y_train, covarianceModel, basis)
fitter_algo.run()
fitter_result = fitter_algo.getResult()
gpr_algo = ot.GaussianProcessRegression(fitter_result)
gpr_algo.run()
gpr_result = gpr_algo.getResult()
gprMetamodel = gpr_result.getMetaModel()

# %%
# We observe that the `scale` and `amplitude` hyper-parameters have been optimized by the :meth:`~openturns.GaussianProcessFitter.run` method.
# The default optimization method (used here) is the :class:`~openturns.Cobyla`, which is different from the old API.
# Then we get the metamodel with `getMetaModel` for evaluating the outputs of the metamodel on the test design of experiments.

# %%
# Now we plot Gaussian process Regression output, in addition to the previous plots

# %%
graph = ot.Graph("Comparison data vs GP models", "", "")
graph.add(plot_1d_data(x_test, y_test, legend="Exact", color="black"))
graph.add(plot_1d_data(x_train, y_train, type="Cloud", legend="Data", color="red"))
graph.add(
    plot_1d_data(
        x_test, krigingMM(x_test), legend="Kriging", color="blue", linestyle="dashed"
    )
)
graph.add(
    plot_1d_data(
        x_test, gprMetamodel(x_test), legend="GPR", color="green", linestyle="dotdash"
    )
)
graph.setAxes(True)
graph.setXTitle("X")
graph.setYTitle("Y")
graph.setLegendPosition("upper right")
view = otv.View(graph)

# %%
# We see that the Gaussian process regression estimated with both classes is interpolating.
# This is what is meant by *conditioning* a Gaussian process.
# We see that, when the sine function has a strong curvature between two points which are separated by a large distance (e.g. between :math:`x=4` and :math:`x=6`),
# then the Gaussian regression is not close to the function :math:`g`.
# However, when the training points are close (e.g. between :math:`x=11` and :math:`x=11.5`) or when the function is nearly linear
# (e.g. between :math:`x=8` and :math:`x=11`),
# then the gaussian process regression is quite accurate.

# %%
# Activating nugget factor
# ------------------------
# Both APIs allow one to estimate the model with an active nugget factor thanks to the :meth:`~openturns.CovarianceModel.activateNuggetFactor`,
# e.g. the parameter is estimated within the optimization process.
#

# %%
covarianceModel.activateNuggetFactor(True)
ot.RandomGenerator.SetSeed(1235)
epsilon = ot.Normal(0, 1.5).getSample(y_train.getSize())

# %%
kriging_algo_wnf = ot.KrigingAlgorithm(
    x_train, y_train + epsilon, covarianceModel, basis
)
kriging_algo_wnf.setOptimizationAlgorithm(ot.NLopt("GN_DIRECT"))
kriging_algo_wnf.run()
kriging_result_wnf = kriging_algo_wnf.getResult()
krigingMM_wnf = kriging_result_wnf.getMetaModel()
print(
    f"Nugget factor estimated with Kriging class = {kriging_result_wnf.getCovarianceModel().getNuggetFactor()}"
)

# %%
fitter_algo_wnf = ot.GaussianProcessFitter(
    x_train, y_train + epsilon, covarianceModel, basis
)
fitter_algo_wnf.setOptimizationAlgorithm(ot.NLopt("GN_DIRECT"))
fitter_algo_wnf.run()
fitter_result_wnf = fitter_algo_wnf.getResult()
gpr_algo_wnf = ot.GaussianProcessRegression(fitter_result_wnf)
gpr_algo_wnf.run()
gpr_result_wnf = gpr_algo_wnf.getResult()
gprMetamodel_wnf = gpr_result_wnf.getMetaModel()
print(
    f"Nugget factor estimated with GPR class = {gpr_result_wnf.getCovarianceModel().getNuggetFactor()}"
)

# %%
# We plot the test and train data
graph = ot.Graph("test and train with noisy data", "", "")
graph.add(plot_1d_data(x_test, y_test, legend="Exact", color="black"))
graph.add(
    plot_1d_data(
        x_train, y_train + epsilon, type="Cloud", legend="Noisy data", color="red"
    )
)
graph.add(
    plot_1d_data(
        x_test,
        krigingMM_wnf(x_test),
        legend="Kriging",
        color="blue",
        linestyle="dashed",
    )
)
graph.add(
    plot_1d_data(
        x_test,
        gprMetamodel_wnf(x_test),
        legend="GPR",
        color="green",
        linestyle="dotdash",
    )
)
graph.setAxes(True)
graph.setAxes(True)
graph.setXTitle("X")
graph.setYTitle("Y")
graph.setLegendPosition("upper right")
view = otv.View(graph)

# %%
# Compute confidence bounds
# -------------------------

# %%
# In order to assess the quality of the surrogate model, we can estimate the variance and compute a 95% confidence interval
# associated with the conditioned Gaussian process.
# We begin by defining the `alpha` variable containing the complementary of the confidence level than we want to compute.
# Then we compute the quantile of the Gaussian distribution corresponding to `1-alpha/2`. Therefore, the confidence interval is:
#
# .. math::
#    P\in\left(X\in\left[q_{\alpha/2},q_{1-\alpha/2}\right]\right)=1-\alpha.
#
#

# %%
alpha = 0.05
quantileAlpha = computeQuantileAlpha(alpha)
print("alpha=%f" % (alpha))
print("Quantile alpha=%f" % (quantileAlpha))

# %%
# In order to compute the regression error, we can consider the conditional variance.
# Within the old API, the `KrigingResult.getConditionalMarginalVariance` method returns the marginal variance `marVar`
# evaluated at each points in the given sample.
# Then we can apply the sqrt function to get the standard deviation.
# Notice that some coefficients in the diagonal are very close to zero and
# nonpositive, which might lead to an exception when applying the sqrt function.
# This is why we add an epsilon on the diagonal, which prevents this issue.

# %%
sqrt = ot.SymbolicFunction(["x"], ["sqrt(x)"])
epsilon = ot.Sample(n_test, [1.0e-8])
conditional_variance_kriging = (
    kriging_result.getConditionalMarginalVariance(x_test) + epsilon
)
conditional_sigma_kriging = sqrt(conditional_variance_kriging)

# %%
# Within the new API, the :meth:`~openturns.GaussianProcessConditionalCovariance.getConditionalMarginalVariance` applies
# and returns the marginal variance `marVar`
# Since this is a variance, we use the square root in order to compute the
# standard deviation.
# Notice also that :meth:`~openturns.GaussianProcessConditionalCovariance.getConditionalCovariance` is similar to
# `KrigingResult.getConditionalCovariance`, and :meth:`~openturns.GaussianProcessConditionalCovariance.getDiagonalCovarianceCollection`
# has a "twin" method `KrigingResult.getConditionalMarginalCovariance`.,

# %%
gccc = ot.GaussianProcessConditionalCovariance(gpr_result)
conditional_variance_gpr = gccc.getConditionalMarginalVariance(x_test)
conditional_sigma_gpr = sqrt(conditional_variance_gpr)

# %%
# Let us compute the same conditional standard deviation when accounting for the noise.

# %%
conditional_variance_kriging_wnf = (
    kriging_result_wnf.getConditionalMarginalVariance(x_test) + epsilon
)
conditional_sigma_kriging_wnf = sqrt(conditional_variance_kriging_wnf)

gccc_wnf = ot.GaussianProcessConditionalCovariance(gpr_result_wnf)
conditional_variance_gpr_wnf = gccc_wnf.getConditionalMarginalVariance(x_test) + epsilon
conditional_sigma_gpr_wnf = sqrt(conditional_variance_gpr_wnf)

# %%
# The following figure presents the conditional standard deviation depending on :math:`x`.

# %%
graph = ot.Graph(
    "Conditional standard deviation", "x", "Conditional standard deviation"
)
curve = ot.Curve(x_test, conditional_sigma_kriging)
graph.add(curve)
curve = ot.Curve(x_test, conditional_sigma_gpr)
graph.add(curve)
graph.setColors(["blue", "red"])
graph.setLegends(["kriging", "GPR"])
graph.setLegendPosition("upper right")
view = otv.View(graph)

# %%
# Select the green colors using HSV values (for the confidence interval)
mycolors = [120, 1.0, 1.0]

# %%
# We are ready to display all the previous information and the three confidence intervals we want.
# First let us evaluate the different confidence bounds

# %%
ci_lower_bound_km, ci_upper_bound_km = computeBoundsConfidenceInterval(
    krigingMM(x_test), quantileAlpha, conditional_sigma_kriging
)
ci_lower_bound_km_noise, ci_upper_bound_km_noise = computeBoundsConfidenceInterval(
    krigingMM_wnf(x_test), quantileAlpha, conditional_sigma_kriging_wnf
)
ci_lower_bound_gpr, ci_upper_bound_gpr = computeBoundsConfidenceInterval(
    gprMetamodel(x_test), quantileAlpha, conditional_sigma_gpr
)
ci_lower_bound_gpr_noise, ci_upper_bound_gpr_noise = computeBoundsConfidenceInterval(
    gprMetamodel_wnf(x_test), quantileAlpha, conditional_sigma_gpr_wnf
)

# %%
# Now we loop over the different models

# %%
grid_layout = ot.GridLayout(2, 2)
grid_layout.setTitle("Confidence interval with various models")
graph = ot.Graph("Kriging API", "x", "y")
boundsPoly = ot.Polygon.FillBetween(x_test, ci_lower_bound_km, ci_upper_bound_km)
boundsPoly.setColor(ot.Drawable.ConvertFromHSV(mycolors[0], mycolors[1], mycolors[2]))
boundsPoly.setLegend(" %d%% bounds" % ((1.0 - alpha) * 100))
graph.add(boundsPoly)
graph.add(
    plot_1d_data(x_test, y_test, legend="Exact", color="black", linestyle="dashed")
)
graph.add(plot_1d_data(x_train, y_train, type="Cloud", legend="Data", color="red"))
graph.add(plot_1d_data(x_test, krigingMM(x_test), legend="Kriging", color="blue"))

graph.setAxes(True)
graph.setXTitle("X")
graph.setYTitle("Y")
graph.setLegendPosition("upper right")
grid_layout.setGraph(0, 0, graph)

# %%
# Gaussian Process Regression

# %%
graph = ot.Graph("GPR API", "x", "y")
boundsPoly = ot.Polygon.FillBetween(x_test, ci_lower_bound_gpr, ci_upper_bound_gpr)
boundsPoly.setColor(ot.Drawable.ConvertFromHSV(mycolors[0], mycolors[1], mycolors[2]))
boundsPoly.setLegend(" %d%% bounds" % ((1.0 - alpha) * 100))
graph.add(boundsPoly)
graph.add(
    plot_1d_data(x_test, y_test, legend="Exact", color="black", linestyle="dashed")
)
graph.add(plot_1d_data(x_train, y_train, type="Cloud", legend="Data", color="red"))
graph.add(plot_1d_data(x_test, gprMetamodel(x_test), legend="GPR", color="blue"))

graph.setAxes(True)
graph.setXTitle("X")
graph.setYTitle("Y")
graph.setLegendPosition("upper right")
grid_layout.setGraph(0, 1, graph)

# %%
# Kriging with noise (old API)

# %%
graph = ot.Graph("Kriging API", "x", "y")
boundsPoly = ot.Polygon.FillBetween(
    x_test, ci_lower_bound_km_noise, ci_upper_bound_km_noise
)
boundsPoly.setColor(ot.Drawable.ConvertFromHSV(mycolors[0], mycolors[1], mycolors[2]))
boundsPoly.setLegend(" %d%% bounds" % ((1.0 - alpha) * 100))
graph.add(boundsPoly)
graph.add(plot_1d_data(x_test, y_test, legend="Exact", color="black"))
graph.add(plot_1d_data(x_train, y_train, type="Cloud", legend="Data", color="red"))
graph.add(
    plot_1d_data(
        x_test,
        krigingMM_wnf(x_test),
        legend="Kriging + noise",
        color="blue",
        linestyle="dashed",
    )
)

graph.setAxes(True)
graph.setXTitle("X")
graph.setYTitle("Y")
graph.setLegendPosition("upper right")
grid_layout.setGraph(1, 0, graph)

# %%
# Gaussian Process Regression with noise

# %%
graph = ot.Graph("GPR API", "x", "y")
boundsPoly = ot.Polygon.FillBetween(
    x_test, ci_lower_bound_gpr_noise, ci_upper_bound_gpr_noise
)
boundsPoly.setColor(ot.Drawable.ConvertFromHSV(mycolors[0], mycolors[1], mycolors[2]))
boundsPoly.setLegend(" %d%% bounds" % ((1.0 - alpha) * 100))
graph.add(boundsPoly)
graph.add(
    plot_1d_data(x_test, y_test, legend="Exact", color="black", linestyle="dashed")
)
graph.add(plot_1d_data(x_train, y_train, type="Cloud", legend="Data", color="red"))
graph.add(
    plot_1d_data(x_test, gprMetamodel_wnf(x_test), legend="GPR + noise", color="blue")
)

graph.setAxes(True)
graph.setXTitle("X")
graph.setYTitle("Y")
graph.setLegendPosition("upper right")
grid_layout.setGraph(1, 1, graph)

view = otv.View(grid_layout)

# %%
# We see that the confidence intervals are small in the regions where two
# consecutive training points are close to each other.
# With noisy data, the confidence interval become bigger.

# %%
# Gaussian Process Regression with fixed trend
# --------------------------------------------

# %%
# The new Gaussian Process Regression allows one to estimate a conditioned Gaussian process regression
# if covariance models are fixed and with a given trend function. Here after how it applies for our use-case.
# First we set the known parameters (covariance, trend)

# %%
scale = [4.51669]
amplitude = [8.648]
covariance_opt = ot.MaternModel(scale, amplitude, 1.5)
trend_function = ot.SymbolicFunction("x", "-3.1710410094572903")

# %%
# Then we define the Gaussian Process Regression relying on these parameters:

# %%
gpr_algo_noopt = ot.GaussianProcessRegression(
    x_train, y_train, covariance_opt, trend_function
)
gpr_algo_noopt.run()
gpr_result_no_opt = gpr_algo_noopt.getResult()
gpr_nopt_Metamodel = gpr_result_no_opt.getMetaModel()

# %%
# Plot the function

# %%
graph = ot.Graph("GPR with known trend", "", "")
graph.add(
    plot_1d_data(x_test, y_test, legend="Exact", color="black", linestyle="dashed")
)
graph.add(plot_1d_data(x_train, y_train, type="Cloud", legend="Data", color="red"))
graph.add(plot_1d_data(x_test, gpr_nopt_Metamodel(x_test), legend="GPR", color="green"))
graph.setAxes(True)
graph.setXTitle("X")
graph.setYTitle("Y")
graph.setLegendPosition("upper right")
view = otv.View(graph)

# %%
# The given GPR matches with the data as expected !

# %%
# -------------------
# Summary of features
# -------------------

# %%
# We illustrated some the features of both old/new API, making a comparison in terms of usage and result.
# We can summarize the main differences hereafter (old API / new API):
#
# * Default optimization solver : :class:`~openturns.TNC`/:class:`~openturns.Cobyla`
# * Conditional covariance : `KrigingResult.getConditionalCovariance`/ :meth:`~openturns.GaussianProcessConditionalCovariance.getConditionalCovariance`
# * Known trend : no / yes (see : :class:`~openturns.GaussianProcessRegression` )
# * Nugget factor : yes / yes
# * Fit the model : `KrigingAlgorithm.run` / :meth:`~openturns.GaussianProcessFitter.run` + :meth:`~openturns.GaussianProcessRegression.run`

# %%
# Display all figures
otv.View.ShowAll()



"""
Gaussian process regression: draw the likelihood
================================================
"""

# %%
# Abstract
# --------
#
# In this short example we draw the log-likelihood as a function of the scale
# parameter of a covariance model. Refer to :ref:`gaussian_process_regression`.
import openturns as ot
import openturns.viewer as otv


# %%
# We define the exact model with a :class:`~openturns.SymbolicFunction` :
f = ot.SymbolicFunction(["x"], ["x*sin(x)"])

# %%
# We use the following input and output training samples :
inputSample = ot.Sample([[1.0], [3.0], [5.0], [6.0], [7.0], [8.0]])
outputSample = f(inputSample)

# %%
# We choose a constant basis for the trend of the metamodel :
basis = ot.ConstantBasisFactory().build()
covarianceModel = ot.SquaredExponential(1)

# %%
# For the covariance model, we use a Matern model with :math:`\nu = 1.5` :
covarianceModel = ot.MaternModel([1.0], 1.5)

# %%
# We are now ready to fit the Gaussian Process parameters and store the result :
fitter = ot.GaussianProcessFitter(inputSample, outputSample, covarianceModel, basis)
fitter.run()
fitter_result = fitter.getResult()


# %%
# We can retrieve the covariance model from the result object and then access
# the scale of the model :
theta = fitter_result.getCovarianceModel().getScale()
print("Scale of the covariance model : %.3e" % theta[0])

# %%
# This hyperparameter is calibrated thanks to a maximization of the log-likelihood. We get this log-likehood as a function of :math:`\theta` :
ot.ResourceMap.SetAsBool("GaussianProcessFitter-UseAnalyticalAmplitudeEstimate", True)
reducedLogLikelihoodFunction = fitter.getReducedLogLikelihoodFunction()

# %%
# We draw the reduced log-likelihood :math:`\mathcal{L}(\theta)` as a function
# of the parameter :math:`\theta`.
graph = reducedLogLikelihoodFunction.draw(0.1, 10.0, 100)
graph.setXTitle(r"$\theta$")
graph.setYTitle(r"$\mathcal{L}(\theta)$")
graph.setTitle(r"Log-likelihood as a function of $\theta$")

# %%
# We represent the estimated parameter as a point on the log-likelihood curve :
L_theta = reducedLogLikelihoodFunction(theta)
cloud = ot.Cloud(theta, L_theta)
cloud.setColor("red")
cloud.setPointStyle("fsquare")
graph.add(cloud)
graph.setLegends([r"Matern $\nu = 1.5$", r"$\theta$ estimate"])

# %%
# We verify on the previous graph that the estimate of :math:`\theta` maximizes
# the log-likelihood.

# %%
# Display figures
view = otv.View(graph)
otv.View.ShowAll()



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()


# We create a numerical math function */
myFunction = ot.SymbolicFunction(["E", "F", "L", "I"], ["-F*L^3/(3*E*I)"])

dim = myFunction.getInputDimension()

# We create a normal distribution point of dimension dim
myDistribution = ot.Normal(dim)

# We create a 'usual' RandomVector from the Distribution
vect = ot.RandomVector(myDistribution)

# We create a composite random vector
output = ot.CompositeRandomVector(myFunction, vect)

# We create an StandardEvent from this RandomVector
myStandardEvent = ot.StandardEvent(output, ot.Less(), 1.0)
print("myStandardEvent=", myStandardEvent)

# We compute one realization of the event
# E = (Y=f(X), operator, threshold)
# E as a RandomVector : Y
print(
    "myStandardEvent (as a RandomVector) realization =",
    repr(ot.RandomVector.getRealization(myStandardEvent)),
)

# E as a Bernoulli
print("myStandardEvent realization=", repr(myStandardEvent.getRealization()))

# We compute a sample of the event
print("myStandardEvent sample=", repr(myStandardEvent.getSample(10)))

# Build a standard event based on an event

R = ot.CorrelationMatrix(dim)
for i in range(dim - 1):
    R[i + 1, i] = 0.5

mean = ot.Point(dim, 0.0)
sigma = ot.Point(dim, 1.0)
myDistribution2 = ot.Normal(mean, sigma, R)

# We create a 'usual' RandomVector from the Distribution
vect2 = ot.RandomVector(myDistribution2)

# We create a composite random vector
output2 = ot.CompositeRandomVector(myFunction, vect2)

# We create an Event from this RandomVector */
myEvent = ot.ThresholdEvent(output2, ot.Less(), 1.0)

# Create a StandardEvent based on this Event */
stdEvent = ot.StandardEvent(myEvent)

# Check if the StandardEvent is really a StandardEvent */
# Get a sample from the second antecedent of the standard event */
size = 2000
# Check if the failure probabilities are the same */
print("Failure probability (Event)=%.6f" % myEvent.getSample(size).computeMean()[0])
print(
    "Failure probability (StandardEvent)=%.6f"
    % stdEvent.getSample(size).computeMean()[0]
)

x = [[1.0, 0.5, 1.0, 0.5], [2.0, 1.0, 2.0, 1.0], [3.0, 1.5, 3.0, 1.5]]
p = [0.3, 0.325, 0.375]
myDistribution3 = ot.FiniteDiscreteDistribution(x, p)

# We create a 'usual' RandomVector from the Distribution */
vect3 = ot.RandomVector(myDistribution3)

# We create a composite random vector */
output3 = ot.CompositeRandomVector(myFunction, vect3)

# We try to create a StandardEvent from this RandomVector */
try:
    myStandardEvent3 = ot.StandardEvent(output3, ot.Less(), 1.0)
except Exception:
    print("Error trying to build myStandardEvent3")



"""
Generate random variates by inverting the CDF
=============================================
"""

# %%
#
# Abstract
# --------
#
# In this example, we show how to generate random variates by inversion of the cumulative distribution function.
# In simple situations, this method is rarely used in practice because of problems with performance,
# statistical quality of the generated random variates and numerical accuracy of the generated numbers when we use floating point numbers.
# However, it is an interesting method to know about because it is a building block for other algorithms and can be used to visualize the distribution of the generated numbers.

# %%
# The WeibullMin distribution
# ---------------------------
#
# Let :math:`\beta>0` and :math:`\alpha>0` be two real parameters: :math:`\alpha` is a shape parameter and  :math:`\beta` is a scale parameter.
#
# The cumulative distribution function of the WeibullMin distribution (also referred as `Weibull distribution`) is:
#
# .. math::
#    F(x) = 1 - \exp\left(-\frac{x-\gamma}{\beta}\right)^\alpha,
#
#
# for any :math:`x\geq \gamma`. For the sake of simplicity, we set :math:`\gamma=0`
#
# In some situations, this parameters are denoted by :math:`k=\beta` and :math:`\lambda=\alpha`.
#
# The inverse of the CDF is:
#
# .. math::
#    F^{-1}(x) = \beta \left(-\log(1-p)\right)^{\frac{1}{\alpha}}
#
#
# for any :math:`x\geq 0`.
#
# This is the *quantile* function, because it computes the quantile :math:`F^{-1}(x)` depending on an outcome :math:`x`.

# %%
# Loss of accuracy when the probability is close to 1
# ---------------------------------------------------
#
# In practice, if the probability :math:`p` is *very* close to 1, then the complementary probability :math:`1-p` is close to zero.
# This can lead to a significant loss of accuracy when we evaluate the subtraction :math:`1-p` with floating point numbers because :math:`p` and 1 have lots of common digits.
# This is called a *loss of accuracy by catastrophic cancellation*, a problem which is common in extreme events.
#
# We can use the :math:`\textrm{expm1}` function, defined by the equation:
#
# .. math::
#    \textrm{expm1}(x) = \exp(x)-1,
#
#
# for any :math:`x\in\mathbb{R}`. This is *not* numerically equivalent to computing `exp` and then subtracting 1.
# Indeed, the `expm1` function is more accurate when its argument `x` is close to zero.
#
# The CDF is then:
#
# .. math::
#    F(x) = -\textrm{expm1} \left(\left(-\frac{x}{\beta}\right)^\alpha\right),
#
#
# for any :math:`x\geq 0`.
#
# For the quantile function, we can use the :math:`\textrm{log1p}` function, defined by the equation:
#
# .. math::
#    \textrm{log1p}(x) = \log(1+x)
#
#
# for any :math:`x>-1`.
#
# Therefore, the quantile function is:
#
# .. math::
#    x = \beta \left(-\textrm{log1p}(-p)\right)^{\frac{1}{\alpha}}
#
#
# for :math:`x\geq 0`.
#
# Note that for :math:`\gamma > 0`, the quantile function writes:
#
# .. math::
#    x = \gamma + \beta \left(-\textrm{log1p}(-p)\right)^{\frac{1}{\alpha}}
#
#
# with :math:`x\geq \gamma`.
#
#
# In the following, we will not use these robust equations and this issue will not be taken into account.

# %%
# Generate by inversion: histogram and density
# --------------------------------------------

# %%
import openturns as ot
import openturns.viewer as otv
import numpy as np

# %%
# The following function defines the quantile function of the :class:`~openturns.WeibullMin` distribution.
# (Of course, we could use the `computeQuantile` method of the :class:`~openturns.WeibullMin` class as well.
# This would create a simpler, but less interesting example: this is a trade off
# that we accept in order to better understand the algorithm.)


def weibullQ(argument):
    """
    WeibullMin quantile function
    """
    p, alpha, beta = argument
    quantile = beta * (-np.log1p(-p)) ** (1 / alpha)
    return [quantile]


# %%
quantileFunction = ot.PythonFunction(3, 1, weibullQ)

# %%
# We define the parameters of the Weibull distribution and create the parametric function.

# %%
alpha = 10.0
beta = 1.0

# %%
quantile = ot.ParametricFunction(quantileFunction, [1, 2], [alpha, beta])
quantile

# %%
# In the library, the uniform distribution is by default over the :math:`[-1,1]` interval. To obtain a uniform distribution over :math:`[0,1]`, we need to set the bounds explicitly.

# %%
U = ot.Uniform(0.0, 1.0)

# %%
# Then we generate a sample of size 1000 from the uniform distribution.

# %%
n = 1000
uniformSample = U.getSample(n)

# %%
# To generate the numbers, we evaluate the quantile function on the uniform numbers.

# %%
weibullSample = quantile(uniformSample)

# %%
# In order to compare the results, we use the :class:`~openturns.WeibullMin` class (using the default value of the location parameter :math:`\gamma=0`).

# %%
W = ot.WeibullMin(beta, alpha)

# %%
histo = ot.HistogramFactory().build(weibullSample).drawPDF()
histo.setTitle("Weibull alpha=%s, beta=%s, n=%d" % (alpha, beta, n))
histo.setLegends(["Sample"])
wpdf = W.drawPDF()
wpdf.setLegends(["Weibull"])
histo.add(wpdf)
view = otv.View(histo)

# %%
# We see that the empirical histogram of the generated outcomes is close to the exact density of the Weibull distribution.

# %%
# Visualization of the quantiles
# ------------------------------

# %%
# We now want to understand the details of the algorithm. To do this, we need to compare the distribution of the uniform numbers with the distribution of the generated points.

# %%
n = 50
uniformSample = U.getSample(n)

# %%
weibullSample = quantile(uniformSample)

# %%
# We sort the sample by increasing order.

# %%
data = ot.Sample(n, 2)
data[:, 0] = weibullSample
data[:, 1] = uniformSample
data.setDescription(["x", "p"])

# %%
sample = ot.Sample(data.sort())
sample[0:5, :]

# %%
weibullSample = sample[:, 0]
uniformSample = sample[:, 1]

# %%
graph = ot.Graph(f"Weibull {alpha=}, {beta=}, {n=}", "x", "U")
# Add the CDF plot
curve = W.drawCDF()
graph.add(curve)
# Plot dashed horizontal & vertical lines
for i in range(n):
    curve = ot.Curve(
        [0.0, weibullSample[i, 0], weibullSample[i, 0]],
        [uniformSample[i, 0], uniformSample[i, 0], 0.0],
    )
    curve.setColor("red")
    curve.setLineStyle("dashed")
    graph.add(curve)
view = otv.View(graph)

# %%
# Show all the graphs.
view.ShowAll()

# %%
# This graphics must be read from the `U` axis on the left to the blue curve (representing the CDF), and down to the `X` axis.
# We see that the horizontal lines on the `U` axis follow the uniform distribution. On the other hand, the vertical lines (on the `X` axis) follow the Weibull distribution.



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()

distribution = ot.Skellam(10.0, 5.0)
size = 10000
sample = distribution.getSample(size)
factory = ot.SkellamFactory()
estimatedDistribution = factory.build(sample)
print("distribution=", repr(distribution))
print("Estimated distribution=", repr(estimatedDistribution))
estimatedDistribution = factory.build()
print("Default distribution=", estimatedDistribution)
estimatedDistribution = factory.build(distribution.getParameter())
print("Distribution from parameters=", estimatedDistribution)
estimatedSkellam = factory.buildAsSkellam(sample)
print("Skellam          =", distribution)
print("Estimated skellam=", estimatedSkellam)
estimatedSkellam = factory.buildAsSkellam()
print("Default skellam=", estimatedSkellam)
estimatedSkellam = factory.buildAsSkellam(distribution.getParameter())
print("Skellam from parameters=", estimatedSkellam)



#! /usr/bin/env python

import openturns as ot
import openturns.experimental as otexp
import openturns.testing as ott
import inspect
from pathlib import Path
import re

ot.TESTPREAMBLE()

# find all instantiable classes
instantiables = []
functions = []
for mod in [ot, otexp, ott]:
    for name, obj in inspect.getmembers(mod):
        if inspect.isclass(obj):
            cn = obj.__name__
            if "_" in cn:
                continue
            try:
                instance = obj()
                instantiables.append(obj)
            except Exception:
                pass
        elif inspect.isfunction(obj):
            cn = obj.__name__
            if "_" not in cn:
                functions.append(obj)

# find missing docstrings
count_class = 0
count_class_undoc = 0
count_methods = 0
count_methods_undoc = 0
for class_ in instantiables:
    count_class += 1
    if class_.__doc__ is None:
        print(f"{class_.__name__} class")
        count_class_undoc += 1
    for attr_name in dir(class_):
        if attr_name[0] == "_":
            continue
        class_attr = getattr(class_, attr_name)
        if callable(class_attr):
            count_methods += 1
            if class_attr.__doc__ is None:
                print(f"{class_.__name__}.{attr_name} method")
                count_methods_undoc += 1

# find all static functions
for name, mod in inspect.getmembers(ot):
    if inspect.ismodule(mod):
        modn = mod.__name__
        if not modn.startswith("openturns") or modn == "openturns":
            continue
        if "_" in modn:
            continue

        for attr_name in dir(mod):
            if "_" in attr_name:
                continue
            obj = getattr(mod, attr_name)

            if not inspect.isclass(obj) and callable(obj):
                symboln = f"{modn.split('.')[1]}.{attr_name}"
                count_methods += 1
                if obj.__doc__ is None:
                    count_methods_undoc += 1
                    print(f"{symboln} method")

count_functions = len(functions)
count_functions_undoc = 0
for function in functions:
    if function.__doc__ is None:
        print(f"{function.__name__} function")
        count_functions_undoc += 1

print(
    f"-- undocumented classes: {count_class_undoc} ({100.0 * count_class_undoc / count_class:.2f}%) --"
)
print(
    f"-- undocumented methods: {count_methods_undoc} ({100.0 * count_methods_undoc / count_methods:.2f}%) --"
)
print(
    f"-- undocumented functions: {count_functions_undoc} ({100.0 * count_functions_undoc / count_functions:.2f}%) --"
)
if count_class_undoc + count_methods_undoc + count_functions_undoc > 110:
    raise ValueError(
        f"too many undocumented class/methods ({count_class_undoc + count_methods_undoc + count_functions_undoc})"
    )

# count extra docstrings methods
count_methods_extra = 0
for swig_file in Path(__file__).parents[1].joinpath("src").glob("*_doc.i"):
    with open(swig_file, encoding="utf-8") as f:
        for line in f.read().splitlines():
            match = re.search(r'%feature\("docstring"\) OT::([\w]*)::([\w]*)', line)
            if match is not None:
                cn = match.group(1)
                mn = match.group(2)
                try:
                    cls = getattr(ot, cn)
                    obj = cls()
                    if not hasattr(obj, mn):
                        print(f"extra {cn}::{mn} method")
                        count_methods_extra += 1
                except Exception:
                    pass
print(f"-- extra method docstrings: {count_methods_extra}")
if count_methods_extra > 10:
    raise ValueError(f"too many extra method docstrings ({count_methods_extra})")



"""Probabilistic distributions submodule."""

# flake8: noqa

# it's split across several modules to speed-up compilation time
from .dist_bundle1 import *
from .dist_bundle2 import *
from .dist_bundle3 import *



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()
ot.RandomGenerator.SetSeed(0)

# Generate the data for the polygonArrays to be drawn
generator = ot.Normal(2)
size = 5
array = []
palette = ot.Drawable.BuildDefaultPalette(size)
for i in range(size):
    vertices = generator.getSample(3)
    array.append(ot.Polygon(vertices, palette[i], palette[size - i - 1]))

graph = ot.Graph("An array of polygons", "x", "y")
graph.add(ot.PolygonArray(array))



#!/usr/bin/env python

import openturns as ot
import openturns.testing as ott

ot.TESTPREAMBLE()
# ot.Log.Show(ot.Log.INFO)

# Time grid parameters
T = 3.0
NT = 32
tg = ot.RegularGrid(0.0, T / NT, NT)

# Toy function to link input processes to the output process
in_dim = 4
out_dim = 1
spatial_dim = 1


class pyf2p(ot.OpenTURNSPythonFieldToPointFunction):
    def __init__(self, mesh):
        super(pyf2p, self).__init__(mesh, in_dim, out_dim)
        self.setInputDescription(["x1", "x2", "x3", "x4"])
        self.setOutputDescription(["g"])

    def _exec(self, X):
        Xs = ot.Sample(X)
        x1, x2, x3, x4 = Xs.computeMean()
        y = x1 + x2 + x3 - x4 + x1 * x2 - x3 * x4 - 0.1 * x1 * x2 * x3
        return [y]


f = ot.FieldToPointFunction(pyf2p(tg))

# First process: elementary process based on a bivariate random vector
f1 = ot.SymbolicFunction(["t"], ["sin(t)"])
f2 = ot.SymbolicFunction(["t"], ["cos(t)^2"])
myBasis = ot.Basis([f1, f2])
coefDis = ot.Normal([1.0] * 2, [0.6] * 2, ot.CorrelationMatrix(2))
p1 = ot.FunctionalBasisProcess(coefDis, myBasis, tg)

# Second process: smooth Gaussian process
p2 = ot.GaussianProcess(ot.SquaredExponential([1.0], [T / 4.0]), tg)

# Third process: elementary process based on a bivariate random vector
randomParameters = ot.JointDistribution([ot.Uniform(), ot.Normal()])
p3 = ot.FunctionalBasisProcess(
    randomParameters,
    ot.Basis(
        [ot.SymbolicFunction(["t"], ["1", "0"]), ot.SymbolicFunction(["t"], ["0", "1"])]
    ),
)

X = ot.AggregatedProcess([p1, p2, p3])
X.setMesh(tg)

N = 1000
x = X.getSample(N)
y = f(x)

# run algo
algo = ot.FieldToPointFunctionalChaosAlgorithm(x, y)
ot.ResourceMap.SetAsUnsignedInteger("FunctionalChaosAlgorithm-BasisSize", N)
ot.ResourceMap.SetAsBool("FunctionalChaosAlgorithm-Sparse", True)
algo.setThreshold(4e-2)
algo.setRecompress(True)
algo.run()
result = algo.getResult()
bi = result.getInputBlockIndices()
print(f"block indices={bi}")
assert len(bi) == 4, "wrong number of blocks"

# inspect modes
kl_results = result.getInputKLResultCollection()
n_modes = [len(res.getEigenvalues()) for res in kl_results]
print(f"n_modes={n_modes}")
ratios = [res.getSelectionRatio() for res in kl_results]
print(f"ratios={ratios}")

# retrieve Sobol indices
sensitivity = ot.FieldFunctionalChaosSobolIndices(result)
sobol_0 = sensitivity.getFirstOrderIndices()
sobol_0t = sensitivity.getTotalOrderIndices()
print(f"first order={sobol_0}")
print(f"total order={sobol_0t}")
ott.assert_almost_equal(sobol_0, [0.0673166, 0.439809, 0.0946051, 0.278286])
ott.assert_almost_equal(sobol_0t, [0.0908364, 0.463752, 0.191125, 0.374326])
graph = sensitivity.draw()
# ot.Show(graph)

sobol2 = dict(
    [((j, i), sensitivity.getSobolIndex([i, j])) for i in range(4) for j in range(i)]
)
print(sobol2)
ott.assert_almost_equal(sobol2[(0, 1)], 0.530589)

# rerun with block indices
blockIndices = [[0], [1], [2, 3]]
algo.setBlockIndices(blockIndices)
algo.run()
result = algo.getResult()

# retrieve Sobol indices
sensitivity = ot.FieldFunctionalChaosSobolIndices(result)
sobol_0 = sensitivity.getFirstOrderIndices()
sobol_0t = sensitivity.getTotalOrderIndices()
print(f"first order={sobol_0}")
print(f"total order={sobol_0t}")
ott.assert_almost_equal(sobol_0, [0.06653, 0.434889, 0.474891])
ott.assert_almost_equal(sobol_0t, [0.0897988, 0.458579, 0.475371])
graph = sensitivity.draw()
# ot.Show(graph)



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()

# We create a numerical math function
myFunction = ot.SymbolicFunction(("E", "F", "L", "I"), ("-F*L^3/(3.*E*I)",))

dim = myFunction.getInputDimension()
# We create a normal distribution point of dimension 1
mean = ot.Point(dim, 0.0)
# E
mean[0] = 50.0
# F
mean[1] = 1.0
# L
mean[2] = 10.0
# I
mean[3] = 5.0
sigma = ot.Point(dim, 1.0)
R = ot.IdentityMatrix(dim)
myDistribution = ot.Normal(mean, sigma, R)

# We create a 'usual' RandomVector from the Distribution
vect = ot.RandomVector(myDistribution)

# We create a composite random vector */
output = ot.CompositeRandomVector(myFunction, vect)

# We create an Event from this RandomVector */
myEvent = ot.ThresholdEvent(output, ot.Less(), -3)

# TEST ONE : directional sampling from an event (slow and safe strategy by default)
# We create a Directional Sampling algorithm */
myAlgo = ot.DirectionalSampling(myEvent)
myAlgo.setMaximumOuterSampling(250)
myAlgo.setBlockSize(4)
myAlgo.setMaximumCoefficientOfVariation(0.1)

print("DirectionalSampling=", myAlgo)

# Perform the simulation */
myAlgo.run()

# Stream out the result */
print("DirectionalSampling result=", myAlgo.getResult())

# TEST TWO : diectional sampling from an event and a root strategy : Risky And Fast, and a directional sampling strategy :
# We create a Directional Sampling algorithm */
myAlgo2 = ot.DirectionalSampling(myEvent, ot.MediumSafe(), ot.OrthogonalDirection())
myAlgo2.setMaximumOuterSampling(250)
myAlgo2.setBlockSize(4)
myAlgo2.setMaximumCoefficientOfVariation(0.1)

print("DirectionalSampling=", myAlgo2)

# Perform the simulation */
myAlgo2.run()

# Stream out the result */
print("DirectionalSampling result=", myAlgo2.getResult())

# Ticket #778
myFunction = ot.SymbolicFunction("x", "x")

X1 = ot.RandomVector(ot.Uniform(-2, 1.99999))
X2 = ot.RandomVector(ot.Uniform(-2, 2.0))

for X in [X1, X2]:
    vector = ot.CompositeRandomVector(myFunction, X)
    event = ot.ThresholdEvent(vector, ot.GreaterOrEqual(), 0.0)

    print("X:", X.getDistribution())

    myAlgo3 = ot.DirectionalSampling(event)
    n1 = myFunction.getEvaluationCallsNumber()
    myAlgo3.run()
    n2 = myFunction.getEvaluationCallsNumber()
    result = myAlgo3.getResult().getProbabilityEstimate()
    print("p=%.6g (ncalls = %d)" % (result, n2 - n1))



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()

meanPoint = ot.Point(1)
meanPoint[0] = 1.0
sigma = ot.Point(1)
sigma[0] = 3.0
R = ot.CorrelationMatrix(1)
R[0, 0] = 1.0

# Create a collection of distribution
dimension = 2000
print("Creating a composed distribution of dimension ", dimension)
aCollection = ot.DistributionCollection(dimension, ot.Normal(meanPoint, sigma, R))
for i in range(dimension):
    aCollection[i] = ot.Normal(meanPoint, sigma, R)

# Create a a copula
aCopula = ot.IndependentCopula(dimension)

# Instantiate one distribution object
distribution = ot.JointDistribution(aCollection, aCopula)
print("Distribution created.")

# Is this distribution elliptical ?
print("Elliptical = ", distribution.isElliptical())

# Has this distribution an elliptical copula?
print("Elliptical copula = ", distribution.hasEllipticalCopula())

# Has this distribution an independent copula?
print("Independent copula = ", distribution.hasIndependentCopula())

# Test for sampling
size = 10
anotherSample = distribution.getSample(size)

# Define a point
zero = ot.Point(dimension, 0.0)

# Show PDF and CDF of zero point
zeroPDF = distribution.computePDF(zero)
zeroCDF = distribution.computeCDF(zero)
print(" pdf=%.6f" % zeroPDF, " cdf=%.6f" % zeroCDF)

# Get 95% quantile
quantile = distribution.computeQuantile(0.95)
print("Quantile=", repr(quantile))
print("CDF(quantile)=%.6f" % distribution.computeCDF(quantile))

# Extract a 2-D marginal
indices = [1, 0]
print("indices=", repr(indices))
margins = distribution.getMarginal(indices)
print("margins=", repr(margins))
print("margins PDF=%.6f" % margins.computePDF(ot.Point(2)))
print("margins CDF=%.6f" % margins.computeCDF(ot.Point(2)))
quantile = ot.Point(margins.computeQuantile(0.5))
print("margins quantile=", repr(quantile))
print("margins CDF(qantile)=%.6f" % margins.computeCDF(quantile))
print("margins realization=", repr(margins.getRealization()))
sample = margins.getSample(1000)
print("margins sample mean=", repr(sample.computeMean()))
print("margins sample covariance=", repr(sample.computeCovariance()))



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()


# Log.Show( Log.Flags() | Log.INFO )

# Problem parameters
dimension = 8

# Create the GSobol' function
# Reference analytical values
meanTh = 1.0
covTh = 1.0
a = ot.Point(dimension)
a[0] = 1.0
a[1] = 2.0
a[2] = 5.0
a[3] = 10.0
a[4] = 20.0
a[5] = 50.0
a[6] = 100.0
a[7] = 500.0
inputVariables = ot.Description(dimension)
formula = ot.Description(1)
formula[0] = "1.0"
for i in range(dimension):
    covTh = covTh * (1.0 + 1.0 / (3.0 * (1.0 + a[i]) ** 2))
    inputVariables[i] = "xi" + str(i)
    formula[0] = (
        formula[0]
        + " * ((abs(4.0 * xi"
        + str(i)
        + " - 2.0) + "
        + str(a[i])
        + ") / (1.0 + "
        + str(a[i])
        + "))"
    )
covTh = covTh - 1.0

model = ot.SymbolicFunction(inputVariables, formula)

# Create the input distribution
distribution = ot.JointDistribution([ot.Uniform(0.0, 1.0)] * dimension)

# Create the orthogonal basis
q = 0.4
enumerateFunction = ot.HyperbolicAnisotropicEnumerateFunction(dimension, q)
productBasis = ot.OrthogonalProductPolynomialFactory(
    [ot.LegendreFactory()] * dimension, enumerateFunction
)

# design experiment
samplingSize = 75

# build basis
basisSize = enumerateFunction.getStrataCumulatedCardinal(30)

listFittingAlgorithm = list()
listFittingAlgorithm.append(ot.KFold())
listFittingAlgorithm.append(ot.CorrectedLeaveOneOut())

for fittingAlgorithmIndex in range(len(listFittingAlgorithm)):
    fittingAlgorithm = listFittingAlgorithm[fittingAlgorithmIndex]
    adaptiveStrategy = ot.FixedStrategy(productBasis, basisSize)
    lsSelectionFactory = ot.LeastSquaresMetaModelSelectionFactory(
        ot.LARS(), fittingAlgorithm
    )
    projectionStrategy = ot.LeastSquaresStrategy(lsSelectionFactory)
    experiment = ot.LowDiscrepancyExperiment(
        ot.SobolSequence(), distribution, samplingSize
    )
    ot.RandomGenerator.SetSeed(0)
    X = experiment.generate()
    Y = model(X)
    algo = ot.FunctionalChaosAlgorithm(
        X, Y, distribution, adaptiveStrategy, projectionStrategy
    )
    algo.run()

    result = algo.getResult()
    print("coeffs = ", result.getCoefficients())

    isLeastSquaresPCE = result.isLeastSquares()
    assert isLeastSquaresPCE
    involvesModelSelectionPCE = result.involvesModelSelection()
    involvesModelSelection = (
        adaptiveStrategy.involvesModelSelection()
        or projectionStrategy.involvesModelSelection()
    )
    assert involvesModelSelection == involvesModelSelectionPCE



#! /usr/bin/env python

import openturns as ot
import openturns.testing as ott

ot.TESTPREAMBLE()


size = 4
indices = ot.Indices(size, 0)
indices.fill()
print("indices=", repr(indices))
print("are indices valid with bound=", size, "? ", indices.check(size))
print("are indices valid with bound=", size // 2, "? ", indices.check(size // 2))
indices[0] = indices[size - 1]
print("indices after transformation=", repr(indices))
print("are indices valid with bound=", size, "? ", indices.check(size))
print("complement wrt 7=", indices.complement(7))
print("empty complement=", ot.Indices().complement(7))

# slicing
indices2 = ot.Indices((1, 2, 3, 4, 5))
print("indices2[2] = ", indices2[2])
print("indices2[1:3] = ", indices2[1:3])
print("indices2[:] = ", indices2[:])
indices2[1:3] = (88, 99)
print("indices2 = ", indices2)

index = indices2.index(88)
assert indices2[index] == 88, "wrong index"

with ott.assert_raises(TypeError):
    index = indices2.index(77777)

# Check contains
indices = ot.Indices()
assert not (0 in indices), "wrong contains"
assert not (1 in indices), "wrong contains"
indices = ot.Indices((1, 2, 3, 4, 5))
assert 1 in indices, "wrong contains"
assert 5 in indices, "wrong contains"
indices = ot.Indices((3, 5, 7))
assert 3 in indices, "wrong contains"
assert 5 in indices, "wrong contains"
assert 7 in indices, "wrong contains"
assert not (2 in indices), "wrong contains"
assert not (4 in indices), "wrong contains"
assert not (0 in indices), "wrong contains"
assert not (1 in indices), "wrong contains"

# Check norm1() and normInf()
indices = ot.Indices(())
assert indices.norm1() == 0, "wrong norm1"
assert indices.normInf() == 0, "wrong norm1"
indices = ot.Indices((3, 5, 7))
assert indices.norm1() == 15, "wrong norm1"
assert indices.normInf() == 7, "wrong norm1"



#! /usr/bin/env python

import openturns as ot
import openturns.testing as ott

ot.TESTPREAMBLE()


# Instantiate one distribution object
distribution = ot.Multinomial(5, ot.Point(3, 0.25))
print("Distribution ", repr(distribution))
print("Distribution ", distribution)

# Is this distribution elliptical ?
print("Elliptical = ", distribution.isElliptical())

# Is this distribution continuous ?
print("Continuous = ", distribution.isContinuous())

# Test for realization of distribution
oneRealization = distribution.getRealization()
print("oneRealization=", oneRealization)

print("support=\n" + str(distribution.getSupport()))
interval = ot.Interval(
    ot.Point(distribution.getDimension(), 1.0),
    ot.Point(distribution.getDimension(), 3.0),
)
print(
    "support restricted to the interval=\n"
    + str(interval)
    + " gives=\n"
    + str(distribution.getSupport(interval))
)

# Define a point
point = ot.Point(distribution.getDimension(), 1.0)
print("Point= ", repr(point))

# Show PDF and CDF at point
LPDF = distribution.computeLogPDF(point)
print("log pdf=%.6f" % LPDF)
PDF = distribution.computePDF(point)
print("pdf     =%.6f" % PDF)
CDF = distribution.computeCDF(point)
print("cdf=%.5f" % CDF)
proba = distribution.computeProbability(
    ot.Interval(
        [i for i in range(distribution.getDimension())],
        [i + 1.0 for i in range(distribution.getDimension())],
    )
)
print("probability=%.5f" % proba)
quantile = distribution.computeQuantile(0.95)
print("quantile=", repr(quantile))
print("cdf(quantile)= %.6f" % distribution.computeCDF(quantile))
print("entropy=%.6f" % distribution.computeEntropy())

mean = distribution.getMean()
print("mean=", repr(mean))
covariance = distribution.getCovariance()
print("covariance=", repr(covariance))
parameters = distribution.getParametersCollection()
print("parameters=", repr(parameters))
parameter = distribution.getParameter()
print("parameter=", repr(parameter))
print("parameterDesc=", distribution.getParameterDescription())
distribution.setParameter(parameter)

ot.Log.Show(ot.Log.TRACE)
validation = ott.DistributionValidation(distribution)
validation.skipMoments()  # slow
validation.skipCorrelation()  # slow
validation.skipConditional()  # FIXME
validation.run()



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()


ot.ResourceMap.Set("SymbolicParser-Backend", "MuParser")
evaluation = ot.SymbolicEvaluation(["x0", "x1", "x2"], ["y0"], ["x0^2+2*x1+3*x2^3"])
hessian = ot.SymbolicHessian(evaluation)
point = [-1.0, 4.0, -4.0]
print("hessian=", hessian)
print("value at", point, "=", hessian.hessian(point))



"""
Draw the QQ-Plot
================
"""

# %%
# In this example we are going to perform a visual goodness-of-fit test for an 1-d distribution with the QQ plot.

# %%
import openturns as ot
import openturns.viewer as otv


# %%
# Create data
distribution = ot.Gumbel(0.2, 0.5)
sample = distribution.getSample(100)
sample.setDescription(["Sample"])

# %%
# Fit a distribution
distribution = ot.GumbelFactory().build(sample)

# %%
# Draw QQ plot
graph = ot.VisualTest.DrawQQplot(sample, distribution)
view = otv.View(graph)

# %%
# Incorrect proposition
graph = ot.VisualTest.DrawQQplot(sample, ot.WeibullMin())
view = otv.View(graph)

# %%
otv.View.ShowAll()



"""
Parallel coordinates graph as sensitivity tool
==============================================
"""

# %%
# The parallel coordinates graph enables to visualize all the combinations of the input
# variables which lead to a specific range of the output variable.
#
# Let us consider a model :math:`f: \mathbb{R}^n \longrightarrow \mathbb{R}`, where :math:`f(\vect{X}) = Y`.
#
# The graph requires to have an input sample :math:`X_s` and an output sample :math:`Y_s`.
#
# The first figure draws such a graph: each column represents one component
# :math:`X_i` of the input vector :math:`\vect{X}`.
# The last column represents the scalar output variable :math:`Y`.
# For each point :math:`\vect{X}^j`, each component :math:`X_i^j`
# is noted on its respective axe and the last mark is the one which corresponds to
# the associated :math:`Y^j`. A line joins all the marks. Thus, each point of the sample
# corresponds to a particular line on the graph.
#
# The scale of the axes are quantile based: each axe runs between 0 and 1 and each
# value is represented by its quantile with respect to its marginal empirical distribution.
#
# It is interesting to select, among those lines, the ones which correspond to a specific range of the output variable.
# These particular lines are colored differently.
# This specific range is defined in the quantile based scale of :math:`Y` or in its specific scale.
# In that second case, the range is automatically converted into a quantile based scale range.

# %%
import openturns as ot
import openturns.viewer as otv

# %%
# Create data to visualize

# Create the model Y = x1^2 + x2
model = ot.SymbolicFunction(["x1", "x2"], ["x1^2+x2"])

# Create the input distribution and random vector X
myCorMat = ot.CorrelationMatrix(2)
myCorMat[0, 1] = -0.6
inputDist = ot.Normal([0.0, 0.0], myCorMat)
inputDist.setDescription(["X1", "X2"])

inputVector = ot.RandomVector(inputDist)

# Create the output random vector Y=model(X)
output = ot.CompositeRandomVector(model, inputVector)

# Generate the input sample
N = 500
X = inputVector.getSample(N)

# Evaluate the associated output sample
Y = model(X)
Y.setDescription("Y")

print(Y.getMin(), Y.getMax(), Y.computeQuantilePerComponent(0.9))

# %%
# Example 1: value based scale to describe the Y range
minValue = 3.35
maxValue = 20.0
quantileScale = False
graphCobweb = ot.VisualTest.DrawParallelCoordinates(
    X, Y, minValue, maxValue, "red", quantileScale
)
graphCobweb.setLegendPosition("lower right")
view = otv.View(graphCobweb)

# %%
# Example 2: rank based scale to describe the Y range
minValue = 0.9
maxValue = 1.0
quantileScale = True
graphCobweb = ot.VisualTest.DrawParallelCoordinates(
    X, Y, minValue, maxValue, "red", quantileScale
)
graphCobweb.setLegendPosition("lower right")
view = otv.View(graphCobweb)

# %%
otv.View.ShowAll()



"""
Add a trend to a process
========================
"""

# %%
# In this example we are going to add a trend to a process.
#
# The :class:`~openturns.TrendTransform` class enables to create a new process :math:`Y` from a process :math:`X` (no hypothesis on :math:`X` needed):
#
# .. math::
#    Y(\omega, t) = X(\omega, t) + f(t)
#

# %%
import openturns as ot
import openturns.viewer as otv

# %%
# Create a process
grid = ot.RegularGrid(0.0, 0.1, 10)
amplitude = [5.0]
scale = [0.2]
covModel = ot.ExponentialModel(scale, amplitude)
X = ot.GaussianProcess(covModel, grid)

# %%
# Draw a sample
sample = X.getSample(6)
sample.setName("X")
graph = sample.drawMarginal(0)
view = otv.View(graph)

# %%
# Define a trend function
f = ot.SymbolicFunction(["t"], ["30*t"])
fTrend = ot.TrendTransform(f, grid)

# %%
# Add it to the process
Y = ot.CompositeProcess(fTrend, X)
Y.setName("Y")

# %%
# Draw a sample
sample = Y.getSample(6)
sample.setName("Y")
graph = sample.drawMarginal(0)
view = otv.View(graph)

# %%
# Display all figures
otv.View.ShowAll()



import openturns as ot
from matplotlib import pyplot as plt
import openturns.viewer as otv
import math as m

a = -m.pi
b = m.pi
f = ot.SymbolicFunction(["x", "y"], ["1+cos(x)*sin(y)"])
ll = [ot.SymbolicFunction(["x"], [" 2+cos(x)"])]
u = [ot.SymbolicFunction(["x"], ["-2-cos(x)"])]

# Draw the graph of the integrand and the bounds:

g = ot.Graph("IteratedQuadrature example", "x", "y")
g.setLegendPosition("upper right")
g.add(f.draw([a, a], [b, b]))
curve = ll[0].draw(a, b).getDrawable(0)
curve.setLineWidth(2)
curve.setColor("red")
g.add(curve)
curve = u[0].draw(a, b).getDrawable(0)
curve.setLineWidth(2)
curve.setColor("red")
g.add(curve)

# Evaluate the integral with high precision:

Iref = ot.IteratedQuadrature(
    ot.GaussKronrod(100000, 1e-13, ot.GaussKronrodRule(ot.GaussKronrodRule.G11K23))
).integrate(f, a, b, ll, u)

# Evaluate the integral with the default GaussKronrod algorithm:

f = ot.MemoizeFunction(f)
I1 = ot.IteratedQuadrature(ot.GaussKronrod()).integrate(f, a, b, ll, u)
sample1 = f.getInputHistory()
print(
    "I1=",
    I1,
    "#evals=",
    sample1.getSize(),
    "err=",
    abs(100.0 * (1.0 - I1[0] / Iref[0])),
    "%",
)
cloud = ot.Cloud(sample1)
cloud.setPointStyle("fcircle")
cloud.setColor("green")
g.add(cloud)
f.clearHistory()

# Evaluate the integral with the default IteratedQuadrature algorithm:

I2 = ot.IteratedQuadrature().integrate(f, a, b, ll, u)
sample2 = f.getInputHistory()
# print('I2=', I2, '#evals=', sample2.getSize(), \
#        'err=', abs(100.0*(1.0-I2[0]/Iref[0])), '%')
cloud = ot.Cloud(sample2)
cloud.setPointStyle("fcircle")
cloud.setColor("gold")
g.add(cloud)

fig = plt.figure(figsize=(4, 4))
axis = fig.add_subplot(111)
axis.set_xlim(auto=True)
otv.View(g, figure=fig, axes=[axis], add_legend=False)



#!/usr/bin/env python

import openturns as ot
from openturns.testing import assert_almost_equal
from openturns.usecases import ishigami_function

ot.RandomGenerator.SetSeed(0)
im = ishigami_function.IshigamiModel()
input_names = im.distribution.getDescription()
n = 250
sampleX = im.distribution.getSample(n)
sampleY = im.model(sampleX)

ot.ResourceMap.SetAsUnsignedInteger("SobolIndicesAlgorithm-DefaultBootstrapSize", 100)

ot.RandomGenerator.SetSeed(42)  # Set seed for bootstrap interval computation algorithm
mySobol = ot.RankSobolSensitivityAlgorithm(sampleX, sampleY)
indices = mySobol.getFirstOrderIndices()
assert_almost_equal(indices, [0.208654, 0.493591, -0.0669488], 1.0e-4, 1.0e-2)

interval = mySobol.getFirstOrderIndicesInterval()
assert_almost_equal(
    interval.getLowerBound(), [0.107531, 0.406036, -0.195595], 1.0e-4, 1.0e-2
)
assert_almost_equal(
    interval.getUpperBound(), [0.328238, 0.569699, 0.0936075], 1.0e-4, 1.0e-2
)

# Testing that we still get the same results when using setDesign
normal = ot.Normal()
mySobol2 = ot.RankSobolSensitivityAlgorithm(normal.getSample(10), normal.getSample(10))
ot.RandomGenerator.SetSeed(42)
mySobol2.setDesign(sampleX, sampleY, sampleX.getSize())

indices2 = mySobol2.getFirstOrderIndices()
interval2 = mySobol2.getFirstOrderIndicesInterval()

assert_almost_equal(indices2, indices, 0, ot.SpecFunc.ScalarEpsilon)
assert_almost_equal(
    interval2.getLowerBound(), interval.getLowerBound(), 0, ot.SpecFunc.ScalarEpsilon
)
assert_almost_equal(
    interval2.getUpperBound(), interval.getUpperBound(), 0, ot.SpecFunc.ScalarEpsilon
)



"""
Visualize pairs between two samples
===================================
"""

# %%
# In this example we visualize the relationships between the marginal samples of two given samples.
# This is usual when we analyze the relationship of output sample of a model depending on input sample.
# This can be achieved by plotting the outputs versus the inputs.
# When there are several inputs (which is the most often encountered case) and several outputs (which is less often).
# The :meth:`~openturns.VisualTest.DrawPairsXY` method provides a tool to plot the pairs of input and output marginal samples and see the correlations graphically.

# %%
import openturns as ot
import openturns.viewer as otv


# %%
# Create the model
model = ot.SymbolicFunction(
    ["X0", "X1", "X2"],
    ["1.0 + 2.0 * X0 - 1.0 * X1 + 4.0 * X2", "-2.0 - 3.0 * X0 + 5.0 * X1 - 1.0 * X2"],
)

# %%
# Create the input data to visualize
distribution = ot.Normal(3)
distribution.setDescription(model.getInputDescription())
input_sample = distribution.getSample(100)

# %% Create the output data to visualize
output_sample = model(input_sample)

# %% Visualize the data
graph = ot.VisualTest.DrawPairsXY(input_sample, output_sample)
graph.setTitle("Clouds of input / output samples")
view = otv.View(graph)
otv.View.ShowAll()



#!/usr/bin/env python

import openturns as ot
import scipy.stats as st
import numpy as np

for scipy_dist in [
    st.uniform(-1.0, 4.0),
    st.johnsonsu(2.55, 2.25),
    st.binom(10, 0.5),
    st.poisson(0.6),
]:
    np.random.seed(42)

    # create an openturns distribution
    py_dist = ot.SciPyDistribution(scipy_dist)
    distribution = ot.Distribution(py_dist)

    print("distribution=", distribution)
    print("continuous?", distribution.isContinuous())
    print("discrete?", distribution.isDiscrete())
    print("integral?", distribution.isIntegral())
    print("realization=", distribution.getRealization())
    sample = distribution.getSample(10000)
    print("sample=", sample[0:5])
    print("pdf@0.1= %.6g" % (distribution.computePDF([0.1])))
    point = [0.0]
    print("pdf= %.6g" % distribution.computePDF(point))
    cdf = distribution.computeCDF(point)
    print("cdf= %.6g" % cdf)
    print("quantile=", distribution.computeQuantile(cdf))
    print("quantile (tail)=", distribution.computeQuantile(cdf, True))
    print("scalar quantile=%.6g" % distribution.computeScalarQuantile(cdf))
    print("scalar quantile (tail)=%.6g" % distribution.computeScalarQuantile(cdf, True))
    print("mean=", distribution.getMean())
    print("mean(sampling)=", sample.computeMean())
    print("std=", distribution.getStandardDeviation())
    print("std(sampling)=", sample.computeStandardDeviation())
    print("skewness=", distribution.getSkewness())
    print("skewness(sampling)=", sample.computeSkewness())
    print("kurtosis=", distribution.getKurtosis())
    print("kurtosis(sampling)=", sample.computeKurtosis())
    print("range=", distribution.getRange())
    if distribution.isDiscrete():
        print("support=", distribution.getSupport())
    parameter = distribution.getParameter()
    print("parameter=", distribution.getParameter())
    if len(parameter) > 0:
        parameter[0] = 3.5
    distribution.setParameter(parameter)
    print("parameter=", distribution.getParameter())
    print("parameterDesc=", distribution.getParameterDescription())



#!/usr/bin/env python

import openturns as ot
import openturns.testing as ott
import sys


def progress(percent):
    sys.stderr.write("-- progress=" + str(percent) + "%\n")


def stop():
    sys.stderr.write("-- stop?\n")
    return False


# Definition of objective function
f = ot.SymbolicFunction(["x1", "x2"], ["-(x2 - 2.0) * (x2 - 2.0)"])
bounds = ot.Interval([-1.0, -1.0e19], [1.0, 1.0e19])
g = ot.SymbolicFunction(["x1", "x2"], ["-(x1 * x1 + x2 - 1.0)"])

# Setting up problem
problem = ot.OptimizationProblem(f)
problem.setBounds(bounds)
problem.setEqualityConstraint(g)

algo = ot.Ipopt(problem)
algo.setStartingPoint([0.5, 1.5])
algo.setMaximumCallsNumber(10000)
algo.setProgressCallback(progress)
algo.setStopCallback(stop)

# ot.ResourceMap.AddAsScalar('Ipopt-max_cpu_time', 15.0)

algo.run()
result = algo.getResult()
print(" -- Optimal point = " + result.getOptimalPoint().__str__())
print(" -- Optimal value = " + result.getOptimalValue().__str__())
print(" -- Evaluation number = " + result.getInputSample().getSize().__str__())
ott.assert_almost_equal(result.getOptimalPoint(), [1.0, 0.0], 1e-5, 1e-4)



import openturns as ot
from matplotlib import pyplot as plt
import openturns.viewer as otv

domain = ot.Interval(-1.0, 1.0)
basis = ot.OrthogonalProductFunctionFactory([ot.FourierSeriesFactory()])
coll = [basis.build(i) for i in range(10)]
experiment = ot.GaussProductExperiment(basis.getMeasure(), [20])
mustScale = False
threshold = 0.001
model = ot.AbsoluteExponential([1.0])
algo = ot.KarhunenLoeveQuadratureAlgorithm(
    domain, domain, model, experiment, coll, mustScale, threshold
)
algo.run()
ev = algo.getResult().getEigenvalues()
modes = algo.getResult().getScaledModes()
g = ot.Graph(r"Quadrature approx. of KL expansion for $C(s,t)=e^{-|s-t|}$")
g.setAxes(True)
g.setGrid(True)
g.setXTitle(r"$t$")
g.setYTitle(r"$\sqrt{\lambda_n}\phi_n$")
for mode in modes:
    g.add(mode.draw(-1.0, 1.0, 256))

fig = plt.figure(figsize=(6, 4))
axis = fig.add_subplot(111)
axis.set_xlim(auto=True)
otv.View(g, figure=fig, axes=[axis], add_legend=False)



"""
Validation of a Karhunen-Loeve decomposition
============================================
"""

# %%
#
# In this example we are going to assess a Karhunen-Loeve decomposition
#

# %%
import openturns as ot
import openturns.viewer as otv

# %%
# Create a Gaussian process.
numberOfVertices = 20
interval = ot.Interval(-1.0, 1.0)
mesh = ot.IntervalMesher([numberOfVertices - 1]).build(interval)
covariance = ot.SquaredExponential()
process = ot.GaussianProcess(covariance, mesh)

# %%
# Decompose it using KL-SVD.
sampleSize = 100
processSample = process.getSample(sampleSize)
threshold = 1.0e-7
algo = ot.KarhunenLoeveSVDAlgorithm(processSample, threshold)
algo.run()
klresult = algo.getResult()

# %%
# Instantiate the validation service.
validation = ot.KarhunenLoeveValidation(processSample, klresult)

# %%
# Plot the residual field.
residualProcessSample = validation.computeResidual()
view = otv.View(residualProcessSample.drawMarginal(0))

# %%
# Plot the residual mean field.
residualMean = validation.computeResidualMean()
view = otv.View(residualMean.drawMarginal(0))

# %%
# Plot the residual standard deviation field.
residualSigmaField = validation.computeResidualStandardDeviation()
view = otv.View(residualSigmaField.drawMarginal(0))

# %%
# Build the validation graph.
view = otv.View(validation.drawValidation())

# %%
# Build the weight graph.
view = otv.View(validation.drawObservationWeight(0))

# %%
# Build the quality graph.
view = otv.View(validation.drawObservationQuality())

# %%
# Display all figures
otv.View.ShowAll()



#! /usr/bin/env python

import openturns as ot
from math import fabs
import openturns.testing as ott

ot.TESTPREAMBLE()


continuousDistributionCollection = ot.DistributionCollection()
discreteDistributionCollection = ot.DistributionCollection()
distributionCollection = ot.DistributionCollection()

beta = ot.Beta(2.0, 1.0, 0.0, 1.0)
distributionCollection.add(beta)
continuousDistributionCollection.add(beta)

gamma = ot.Gamma(1.0, 2.0, 3.0)
distributionCollection.add(gamma)
continuousDistributionCollection.add(gamma)

gumbel = ot.Gumbel(1.0, 2.0)
distributionCollection.add(gumbel)
continuousDistributionCollection.add(gumbel)

lognormal = ot.LogNormal(1.0, 1.0, 2.0)
distributionCollection.add(lognormal)
continuousDistributionCollection.add(lognormal)

logistic = ot.Logistic(1.0, 1.0)
distributionCollection.add(logistic)
continuousDistributionCollection.add(logistic)

normal = ot.Normal(1.0, 2.0)
distributionCollection.add(normal)
continuousDistributionCollection.add(normal)

truncatednormal = ot.TruncatedNormal(1.0, 1.0, 0.0, 3.0)
distributionCollection.add(truncatednormal)
continuousDistributionCollection.add(truncatednormal)

student = ot.Student(10.0, 10.0, 1.0)
distributionCollection.add(student)
continuousDistributionCollection.add(student)

triangular = ot.Triangular(-1.0, 2.0, 4.0)
distributionCollection.add(triangular)
continuousDistributionCollection.add(triangular)

uniform = ot.Uniform(1.0, 2.0)
distributionCollection.add(uniform)
continuousDistributionCollection.add(uniform)

weibull = ot.WeibullMin(1.0, 1.0, 2.0)
distributionCollection.add(weibull)
continuousDistributionCollection.add(weibull)

geometric = ot.Geometric(0.5)
distributionCollection.add(geometric)
discreteDistributionCollection.add(geometric)

binomial = ot.Binomial(10, 0.25)
distributionCollection.add(binomial)
discreteDistributionCollection.add(binomial)

zipf = ot.ZipfMandelbrot(20, 5.25, 2.5)
distributionCollection.add(zipf)
discreteDistributionCollection.add(zipf)

poisson = ot.Poisson(5.0)
distributionCollection.add(poisson)
discreteDistributionCollection.add(poisson)

x = [[1.0], [2.0], [3.0]]
p = [0.3, 0.2, 0.5]
userdefined = ot.FiniteDiscreteDistribution(x, p)
distributionCollection.add(userdefined)
discreteDistributionCollection.add(userdefined)

size = 100

# Number of continuous distributions
continuousDistributionNumber = continuousDistributionCollection.getSize()
# Number of discrete distributions
discreteDistributionNumber = discreteDistributionCollection.getSize()
# Number of distributions
distributionNumber = continuousDistributionNumber + discreteDistributionNumber

# We create a collection of Sample of size "size" and of
# dimension 1 (scalar values) : the collection has distributionNumber
# Samples

sampleCollection = [ot.Sample(size, 1) for i in range(distributionNumber)]
# We create a collection of Sample of size "size" and of
# dimension 1 (scalar values) : the collection has
# continuousDistributionNumber Samples
continuousSampleCollection = [
    ot.Sample(size, 1) for i in range(continuousDistributionNumber)
]
# We create a collection of Sample of size "size" and of
# dimension 1 (scalar values) : the collection has
# discreteDistributionNumber Samples
discreteSampleCollection = [
    ot.Sample(size, 1) for i in range(discreteDistributionNumber)
]

ot.RandomGenerator.SetSeed(0)
for i in range(continuousDistributionNumber):
    continuousSampleCollection[i] = continuousDistributionCollection[i].getSample(size)
    continuousSampleCollection[i].setName(continuousDistributionCollection[i].getName())
    sampleCollection[i] = continuousSampleCollection[i]
for i in range(discreteDistributionNumber):
    discreteSampleCollection[i] = discreteDistributionCollection[i].getSample(size)
    discreteSampleCollection[i].setName(discreteDistributionCollection[i].getName())
    sampleCollection[continuousDistributionNumber + i] = discreteSampleCollection[i]

factoryCollection = [
    ot.UniformFactory(),
    ot.BetaFactory(),
    ot.NormalFactory(),
    ot.KernelSmoothing(),
]
ot.RandomGenerator.SetSeed(0)
aSample = ot.Uniform(-1.5, 2.5).getSample(size)
model, best_bic = ot.FittingTest.BestModelBIC(aSample, factoryCollection)
print("best model BIC=", repr(model))
model, best_result = ot.FittingTest.BestModelLilliefors(aSample, factoryCollection)
model, best_AIC = ot.FittingTest.BestModelAIC(aSample, factoryCollection)
print("best model AIC=", repr(model))
model, best_AICc = ot.FittingTest.BestModelAICC(aSample, factoryCollection)
print("best model AICc=", repr(model))
print("best model Lilliefors=", repr(model))

# BIC adequation
resultBIC = ot.SquareMatrix(distributionNumber)
for i in range(distributionNumber):
    for j in range(distributionNumber):
        value = ot.FittingTest.BIC(sampleCollection[i], distributionCollection[j], 0)
        # TODO JM: remove the check after the use of infs has been thoroughly tested
        if value < ot.SpecFunc.Infinity:
            resultBIC[i, j] = value
        else:
            resultBIC[i, j] = value * 2.0
print("resultBIC=", repr(resultBIC))

# Kolmogorov test : case with estimated parameters
print("Lilliefors test : case with estimated parameters")
distribution = ot.Normal()
ot.RandomGenerator.SetSeed(0)
sample = distribution.getSample(30)
factory = ot.NormalFactory()
# ot.ResourceMap.SetAsUnsignedInteger(
#    'FittingTest-LillieforsMaximumSamplingSize', 1000000)
# -ot.ResourceMap.SetAsScalar('FittingTest-LillieforsPrecision', 0.0)
fitted_dist, test_result = ot.FittingTest.Lilliefors(sample, factory)
p_exact = 0.50076  # With a sample size equal to 1000000 and a precision equal to 0
pvalue = test_result.getPValue()
rtol = 0.0
atol = 1.0e-2
ott.assert_almost_equal(pvalue, p_exact, rtol, atol)

# Kolmogorov test : case with known parameters
print("Kolmogorov test : case with known parameters")
# Data from PlantGrowth$weight in the MASS R package,
# except the first value which generates a tie
distribution = ot.Normal()
data = (
    5.58,
    5.18,
    6.11,
    4.50,
    4.61,
    5.17,
    4.53,
    5.33,
    5.14,
    4.81,
    4.17,
    4.41,
    3.59,
    5.87,
    3.83,
    6.03,
    4.89,
    4.32,
    4.69,
    6.31,
    5.12,
    5.54,
    5.50,
    5.37,
    5.29,
    4.92,
    6.15,
    5.80,
    5.26,
)
sample = ot.Sample([[x] for x in data])
mean = sample.computeMean()
sample = sample - mean
test_result = ot.FittingTest.Kolmogorov(sample, distribution)
p_exact = 0.8053771610533257963
pvalue = test_result.getPValue()
ott.assert_almost_equal(pvalue, p_exact)
D = test_result.getStatistic()
D_exact = 0.11393533907737134203
ott.assert_almost_equal(D, D_exact)
quality = test_result.getBinaryQualityMeasure()
assert quality
threshold = test_result.getThreshold()
defaultLevel = 0.05
assert threshold == defaultLevel

# Kolmogorov test : case with known parameters, set the level
print("Kolmogorov test : case with known parameters, set the level")
distribution = ot.Normal()
ot.RandomGenerator.SetSeed(0)
sample = distribution.getSample(30)
level = 0.01
test_result = ot.FittingTest.Kolmogorov(sample, distribution, level)
threshold = test_result.getThreshold()
assert threshold == level

# Kolmogorov adequation
resultKolmogorov = ot.SquareMatrix(continuousDistributionNumber)
for i in range(continuousDistributionNumber):
    for j in range(continuousDistributionNumber):
        sample = continuousSampleCollection[i]
        distribution = continuousDistributionCollection[j]
        test_result = ot.FittingTest.Kolmogorov(sample, distribution, 0.05)
        pvalue = test_result.getPValue()
        if fabs(pvalue) < 1.0e-6:
            value = 0.0
        resultKolmogorov[i, j] = value
print("resultKolmogorov=", repr(resultKolmogorov))

# ChiSquared adequation
resultChiSquared = ot.SquareMatrix(discreteDistributionNumber)
for i in range(discreteDistributionNumber):
    for j in range(discreteDistributionNumber):
        try:
            sample = continuousSampleCollection[i]
            distribution = continuousDistributionCollection[j]
            test_result = ot.FittingTest.ChiSquared(sample, distribution, 0.05, 0)
            pvalue = test_result.getPValue()
            if fabs(pvalue) < 1.0e-6:
                value = 0.0
            resultChiSquared[i, j] = value
        except Exception:
            print(
                "Sample=",
                discreteSampleCollection[i],
                " is not compatible with distribution=",
                discreteDistributionCollection[j],
            )
print("resultChiSquared=", repr(resultChiSquared))
# Example taken from the R documentation of chisq.test
s = [[0.0]] * 89 + [[1.0]] * 37 + [[2.0]] * 30 + [[3.0]] * 28 + [[4.0]] * 2
d = ot.FiniteDiscreteDistribution(
    [[0.0], [1.0], [2.0], [3.0], [4.0]], [0.4, 0.2, 0.2, 0.15, 0.05]
)
print("R example p-value=%.5g" % ot.FittingTest.ChiSquared(s, d).getPValue())



"""
Use case : FireSatellite test function
======================================
"""

import openturns as ot
import math as m


class FireSatelliteModel:
    """
    Data class for the Fire Satellite.

    Attributes
    ----------

    dim : int
        Dimension of the problem, dim = 9

    H : :class:`~openturns.TruncatedNormal`
        Altitude (m) distribution
        First marginal, ot.TruncatedNormal(18e6,1e6,18e6-3e6,18e6+3e6)

    Pother : :class:`~openturns.TruncatedNormal`
        Power other than ACS (W) distribution
        Second marginal, ot.TruncatedNormal(1000.0,50.0,1000.0-150.0,1000.0+150.0)

    Fs : :class:`~openturns.TruncatedNormal`
        Average solar flux (W/m^2) distribution
        Third marginal, ot.TruncatedNormal(1400.0,20.0,1400.0-60.0,1400.0+60.0)

    theta : :class:`~openturns.TruncatedNormal`
        Deviation of moment axis (deg) distribution
        Fourth marginal, ot.TruncatedNormal(15.0,1.0,15.0-3.0,15.0+3.0)

    Lsp : :class:`~openturns.TruncatedNormal`
        Moment arm for radiation torque (m) distribution
        Fifth marginal, ot.TruncatedNormal(2.0,0.4,2.0-1.2,2.0+1.2)

    q : :class:`~openturns.TruncatedNormal`
        Reflectance factor (-) distribution
        Sixth marginal, ot.TruncatedNormal(0.5,0.1,0.5-0.3,0.5+0.3)

    RD : :class:`~openturns.TruncatedNormal`
        Residual dipole of spacecraft (A.m^2) distribution
        Seventh marginal, ot.TruncatedNormal(5.0,1.0,5.0-3.0,5.0+3.0)

    Lalpha : :class:`~openturns.TruncatedNormal`
        Moment arm for aerodynamic torque (m) distribution
        Eighth marginal, ot.TruncatedNormal(2.0,0.4,2.0-1.2,2.0+1.2)

    Cd : :class:`~openturns.TruncatedNormal`
        Drag coefficient (-) distribution
        Ninth marginal, ot.TruncatedNormal(1.0,0.3,1.0-0.9,1.0+0.9)

    distribution : :class:`~openturns.JointDistribution`
        The joint distribution of the input parameters.

    model : :class:`~openturns.PythonFunction`
        The Fire Satellite model with H, Pother, Fs, theta, Lsp, q, RD, Lalpha and Cd as variables.
        This function retrieves three outputs : the total torque, the total power and the area of solar array

    modelTotalTorque : :class:`~openturns.PythonFunction`
        The Fire Satellite model retrieving only the  Total Torque as output, with H, Pother, Fs, theta, Lsp, q, RD, Lalpha and Cd as variables.

    modelTotalPower : :class:`~openturns.PythonFunction`
        The Fire Satellite model retrieving only the  Total Power as output, with H, Pother, Fs, theta, Lsp, q, RD, Lalpha and Cd as variables.
        This function retrieves three outputs : the total torque, the total power and the area of solar array

    modelSolarArrayArea : :class:`~openturns.PythonFunction`
        The Fire Satellite model retrieving only the  Solar Array Area as output, with H, Pother, Fs, theta, Lsp, q, RD, Lalpha and Cd as variables.
        This function retrieves three outputs : the total torque, the total power and the area of solar array

    c : float
        Speed of light, c = 2.9979e8 m/s

    omega_max : float
        Maximum rotational velocity of reaction wheel, omega_max = 6000 rpm

    n : int
        Number of reaction wheels that could be active, n = 3

    delta_theta_slew : float
        Slewing time period, delta_theta_slew = 760 s

    As : float
        Area reflecting radiation, As =  13.85 m^2

    i : float
        Sun incidence angle, i = 0 deg

    M : float
        Magnetic moment of earth, M = 7.96e15 A.m^2

    rho : float
        Atmospheric density, rho = 5.1480e-11 kg/m^3

    A : float
        Cross-sectional in flight direction, A = 13.85 m^2

    Phold : float
        Holding power, Phold = 20 W

    mu : float
        Earth gravity constant, mu = 398600.4418e9 m^3/s^2

    Id : float
        Inherent degradation of array, Id = 0.77

    t : float
        Thickness of solar panels, t = 0.005 m

    n_sa : int
        Number of solar arrays, n_sa = 3

    epsilon_deg : float
        Degradation in power production capability, epsilon_deg = 0.0375 percent per year

    LT : float
        Lifetime of spacecraft, LT = 15 years

    r_lw : float
        Length to width ratio of solar array, r_lw = 3

    D : float
        Distance between panels, D = 2 m

    I_bodyX : float
        Inertia of body, X axis, I_bodyX = 6200 kg.m^2

    I_bodyY : float
        Inertia of body, Y axis, I_bodyY =  6200 kg.m^2

    I_bodyZ : float
        Inertia of body, Z axis, I_bodyZ =  4700 kg.m^2

    rho_sa : float
        Average mass density to arrays, rho_sa =  700 kg.m^3

    eta : float
        Power efficiency, eta =  0.22

    phi_target : float
        Target diameter, phi_target = 235000 m

    RE : float
        Earth radius, RE = 6378140 m

    tolFPI : float
        Tolerance on Fixed Point Iteration used in the multidisciplinary analysis, tolFPI = 1e-3

    maxFPIIter : int
        Maximum number of iterations of Fixed Point Iteration used in the multidisciplinary analysis, maxFPIIter = 50


    Examples
    --------
    >>> from openturns.usecases.fire_satellite import FireSatelliteModel
    >>> # Load the FireSatellite model model
    >>> m = FireSatelliteModel()
    """

    def __init__(self):
        # dimension
        self.dim = 9

        # Altitude
        self.H = ot.TruncatedNormal(18e6, 1e6, 18e6 - 3 * 1e6, 18e6 + 3 * 1e6)
        self.H.setDescription("H")

        # Power other than ACS
        self.Pother = ot.TruncatedNormal(
            1000.0, 50.0, 1000.0 - 3 * 50.0, 1000.0 + 3 * 50.0
        )
        self.Pother.setDescription(["Pother"])

        # Average solar flux
        self.Fs = ot.TruncatedNormal(1400.0, 20.0, 1400.0 - 3 * 20.0, 1400.0 + 3 * 20.0)
        self.Fs.setDescription(["Fs"])

        # Deviation of moment axis
        self.theta = ot.TruncatedNormal(15.0, 1.0, 15.0 - 3 * 1.0, 15.0 + 3 * 1.0)
        self.theta.setDescription(["theta"])

        # Moment arm for radiation torque
        self.Lsp = ot.TruncatedNormal(2.0, 0.4, 2.0 - 3 * 0.4, 2.0 + 3 * 0.4)
        self.Lsp.setDescription(["Lsp"])

        # Reflectance factor
        self.q = ot.TruncatedNormal(0.5, 0.1, 0.5 - 3 * 0.1, 0.5 + 3 * 0.1)
        self.q.setDescription(["q"])

        # Residual dipole of spacecraft
        self.RD = ot.TruncatedNormal(5.0, 1.0, 5.0 - 3 * 1.0, 5.0 + 3 * 1.0)
        self.RD.setDescription(["RD"])

        # Moment arm for aerodynamic torque
        self.Lalpha = ot.TruncatedNormal(2.0, 0.4, 2.0 - 3 * 0.4, 2.0 + 3 * 0.4)
        self.Lalpha.setDescription(["Lalpha"])

        # Drag coefficient
        self.Cd = ot.TruncatedNormal(1.0, 0.3, 1.0 - 3 * 0.3, 1.0 + 3 * 0.3)
        self.Cd.setDescription(["Cd"])

        # Input distribution
        self.distribution = ot.JointDistribution(
            [
                self.H,
                self.Pother,
                self.Fs,
                self.theta,
                self.Lsp,
                self.q,
                self.RD,
                self.Lalpha,
                self.Cd,
            ]
        )
        self.distribution.setDescription(
            ["H", "Pother", "Fs", "theta", "Lsp", "q", "RD", "Lalpha", "Cd"]
        )
        self.inputDistribution = self.distribution  # deprecated

        # Definitions of models
        # Global model
        self.model = ot.PythonFunction(9, 3, self.multidisciplinaryAnalysis)

        # Model of Total torque
        def multidisciplinaryAnalysisToTalTorque(x):
            return [self.multidisciplinaryAnalysis(x)[0]]

        self.modelTotalTorque = ot.PythonFunction(
            9, 1, multidisciplinaryAnalysisToTalTorque
        )
        self.modelTotalTorque.setInputDescription(self.distribution.getDescription())
        self.modelTotalTorque.setOutputDescription(["Total Torque"])

        # Model of Total power
        def multidisciplinaryAnalysisToTalPower(x):
            return [self.multidisciplinaryAnalysis(x)[1]]

        self.modelTotalPower = ot.PythonFunction(
            9, 1, multidisciplinaryAnalysisToTalPower
        )
        self.modelTotalPower.setInputDescription(self.distribution.getDescription())
        self.modelTotalPower.setOutputDescription(["TotalPower"])

        # Model of Solar Array Area
        def multidisciplinaryAnalysisSolarArrayArea(x):
            return [self.multidisciplinaryAnalysis(x)[2]]

        self.modelSolarArrayArea = ot.PythonFunction(
            9, 1, multidisciplinaryAnalysisSolarArrayArea
        )
        self.modelSolarArrayArea.setInputDescription(self.distribution.getDescription())
        self.modelSolarArrayArea.setOutputDescription(["SolarArrayArea"])

        # Optional variables (deterministic)
        # Speed of light
        self.c = 2.9979e8

        # Maximum rotational velocity of reaction wheel
        self.omega_max = 6000.0

        # Number of reaction wheels that could be active
        self.n = int(3)

        # Slewing time period
        self.delta_theta_slew = 760.0

        # Area reflecting radiation
        self.As = 13.85

        # Sun incidence angle
        self.i = 0.0

        # Magnetic moment of earth
        self.M = 7.96e15

        # Atmospheric density
        self.rho = 5.1480e-11

        # Cross-sectional in flight direction
        self.A = 13.85

        # Holding power
        self.Phold = 20.0

        # Earth gravity constant
        self.mu = 3.98600e14

        # Inherent degradation of array
        self.Id = 0.77

        # Thickness of solar panels
        self.t = 0.005

        # Number of solar arrays
        self.n_sa = int(3)

        # Degradation in power production capability
        self.epsilon_deg = 0.0375

        # Lifetime of spacecraft
        self.LT = 15.0

        # Length to width ratio of solar array
        self.r_lw = 3.0

        # Distance between panels
        self.D = 2.0

        # Inertia of body, X axis
        self.I_bodyX = 6200.0

        # Inertia of body, Y axis
        self.I_bodyY = 6200.0

        # Inertia of body, Z axis
        self.I_bodyZ = 4700.0

        # Average mass density to arrays
        self.rho_sa = 700.0

        # Power efficiency
        self.eta = 0.22

        # Target diameter
        self.phi_target = 235000.0

        # Earth radius
        self.RE = 6378140

        # Fixed Point Iteration tolerance
        self.tolFPI = 1e-3

        # Fixed Point Iteration maximum iteration
        self.maxFPIIter = int(50)

    def __getstate__(self):
        state = {}
        for k, v in self.__dict__.items():
            # avoid recursive serialization of PythonFunction
            if isinstance(v, ot.Function):
                continue
            state[k] = v
        return state

    def power(self, inputs):
        """
        Function computing the power discipline outputs to retrieve the inertia, the total power and the area of solar array

        :inputs: dictionary of inputs of the Power discipline

        """
        Fs = inputs["Fs"]
        P_ACS = inputs["P_ACS"]
        P_other = inputs["P_other"]
        delta_t_orbit = inputs["delta_t_orbit"]
        delta_t_eclipse = inputs["delta_t_eclipse"]

        n_sa = self.n_sa
        LT = self.LT
        epsilon_deg = self.epsilon_deg
        D = self.D
        r_lw = self.r_lw
        I_bodyX = self.I_bodyX
        I_bodyY = self.I_bodyY
        I_bodyZ = self.I_bodyZ
        Id = self.Id
        rho_sa = self.rho_sa
        t = self.t
        eta = self.eta

        # Total power
        Ptot = P_ACS + P_other

        # power production capability at beginning of life
        P_BOL = eta * Fs * Id * m.cos(self.i)

        # power production capability at end of life
        P_EOL = P_BOL * (1 - epsilon_deg) ** (LT)

        # spacecraft power requirement for eclipse
        Pe = Ptot

        # spacecraft power requirement for daylight
        Pd = Ptot

        # Time per orbit spent in eclipse
        Te = delta_t_eclipse

        # Time per orbit spent in sunlight
        Td = delta_t_orbit - Te

        # Required power output
        Xe = 0.6
        Xd = 0.8
        P_sa = (Pe * Te / Xe + Pd * Td / Xd) / Td

        # total array size
        A_sa = P_sa / P_EOL

        # Determination of moment of inertia
        L = m.sqrt(A_sa * r_lw / n_sa)
        W = m.sqrt(A_sa / (r_lw * n_sa))
        m_sa = 2 * rho_sa * L * W * t
        I_saX = m_sa * (1 / 12 * (L**2 + t**2) + (D + L / 2) ** 2)
        I_saY = m_sa / 12 * (t**2 + W**2)
        I_saZ = m_sa * (1 / 12 * (L**2 + W**2) + (D + L / 2) ** 2)

        # total moment of inertia
        I_tot = ot.Sample([[I_saX + I_bodyX], [I_saY + I_bodyY], [I_saZ + I_bodyZ]])
        # Minimal inertia
        I_min = I_tot.getMin()[0]
        # Maximal inertia
        I_max = I_tot.getMax()[0]

        outputs = {}
        outputs["Imax"] = I_max
        outputs["Imin"] = I_min
        outputs["A_sa"] = A_sa
        outputs["P_tot"] = Ptot

        return outputs

    def orbit(self, inputs):
        """
        Function computing the orbit discipline outputs and retrieve the slewing angle, the velocity, the orbit duration and the eclipse duration

        :inputs: dictionary of inputs of the Orbit discipline

        """
        mu = self.mu
        phi_target = self.phi_target
        RE = self.RE

        H = inputs["H"]

        # satellite velocity
        v = m.sqrt(mu / (RE + H))
        # orbit period
        delta_t_orbit = 2 * m.pi * (RE + H) / v
        # maximum eclipse time
        delta_t_eclipse = delta_t_orbit / m.pi * m.asin(RE / (RE + H))
        # maximum slewing angle
        theta_slew = m.atan(
            m.sin(phi_target / RE) / (1 - m.cos(phi_target / RE) + H / RE)
        )

        outputs = {}
        outputs["theta_slew"] = theta_slew
        outputs["v"] = v
        outputs["delta_t_orbit"] = delta_t_orbit
        outputs["delta_t_eclipse"] = delta_t_eclipse

        return outputs

    def attitudeControl(self, inputs):
        """
        Function computing the attitude and control discipline outputs to retrieve the power of ACS and total torque

        :inputs: dictionary of inputs of the Attitude and Control discipline

        """
        delta_theta_slew = self.delta_theta_slew
        mu = self.mu
        c = self.c
        As = self.As
        i = self.i
        M = self.M
        rho = self.rho
        A = self.A
        omega_max = self.omega_max
        n = self.n
        Phold = self.Phold
        RE = self.RE

        theta_slew = inputs["theta_slew"]
        Imax = inputs["Imax"]
        Imin = inputs["Imin"]
        H = inputs["H"]
        theta = m.pi / 180.0 * inputs["theta"]
        L_sp = inputs["L_sp"]
        Fs = inputs["Fs"]
        q = inputs["q"]
        R_D = inputs["R_D"]
        C_d = inputs["C_d"]
        L_alpha = inputs["L_alpha"]
        v = inputs["v"]

        # slewing torque
        tau_slew = 4 * theta_slew * Imax / delta_theta_slew**2

        # torque due to gravity gradients
        tau_g = 3 * mu / (2 * RE + H) ** 3 * abs(Imax - Imin) * m.sin(2 * theta)

        # torque due to solar radiation
        tau_sp = L_sp * Fs / c * As * (1 + q) * m.cos(i)

        # torque due to magnetic fiel interactions
        tau_m = 2 * M * R_D / (RE + H) ** 3

        # torque due to atmospheric drag
        tau_alpha = 0.5 * rho * L_alpha * C_d * A * v**2

        # total disturbance torque
        tau_dist = m.sqrt(tau_sp**2 + tau_m**2 + tau_g**2 + tau_alpha**2)

        # total torque
        tau_tot = ot.Sample([[tau_dist], [tau_slew]]).getMax()[0]

        # attitude power control
        P_ACS = tau_tot * omega_max + n * Phold

        outputs = {}
        outputs["P_ACS"] = P_ACS
        outputs["tau_tot"] = tau_tot

        return outputs

    def multidisciplinaryAnalysis(self, x):
        """
        Function computing the multidisciplinary analysis to retrieve the total torque, the total power and the area of solar array

        :x: list of inputs

        """
        H = x[0]
        Pother = x[1]
        Fs = x[2]
        theta = x[3]
        Lsp = x[4]
        q = x[5]
        RD = x[6]
        Lalpha = x[7]
        Cd = x[8]

        tolP_ACS = 1e10

        # run of orbit discipline
        inputs_orbit = {}
        inputs_orbit["H"] = H

        outputs_orbit = self.orbit(inputs_orbit)

        itFPI = 0

        while tolP_ACS > self.tolFPI and itFPI < self.maxFPIIter:
            # run of power discipline
            inputs_power = {}
            inputs_power["theta"] = theta
            inputs_power["Fs"] = Fs
            inputs_power["P_other"] = Pother
            inputs_power["delta_t_orbit"] = outputs_orbit["delta_t_orbit"]
            inputs_power["delta_t_eclipse"] = outputs_orbit["delta_t_eclipse"]

            global outputs_attitude
            if itFPI == 0:
                inputs_power["P_ACS"] = 150.0
            else:
                inputs_power["P_ACS"] = outputs_attitude["P_ACS"]

            outputs_power = self.power(inputs_power)

            # run of attitude discipline
            inputs_attitude = {}
            inputs_attitude["theta_slew"] = outputs_orbit["theta_slew"]
            inputs_attitude["v"] = outputs_orbit["v"]
            inputs_attitude["Imax"] = outputs_power["Imax"]
            inputs_attitude["Imin"] = outputs_power["Imin"]
            inputs_attitude["H"] = H
            inputs_attitude["theta"] = theta
            inputs_attitude["L_sp"] = Lsp
            inputs_attitude["Fs"] = Fs
            inputs_attitude["q"] = q
            inputs_attitude["R_D"] = RD
            inputs_attitude["C_d"] = Cd
            inputs_attitude["L_alpha"] = Lalpha

            outputs_attitude = self.attitudeControl(inputs_attitude)
            tolP_ACS = abs(inputs_power["P_ACS"] - outputs_attitude["P_ACS"])

            itFPI += 1

        return [
            outputs_attitude["tau_tot"],
            outputs_power["P_tot"],
            outputs_power["A_sa"],
        ]



"""
Define a connection function with a field output
================================================
"""

# %%
# In this example, we define a function which has a vector input and a field output.
# The goal of this example is to show how to use :class:`~openturns.PointToFieldConnection` class to combine two functions.
# A detailed explanation of the model is presented :ref:`here <use-case-viscous-fall>`.

# %%
# Define the model
# ----------------

# %%
import openturns as ot
import openturns.viewer as otv
from openturns.usecases import viscous_free_fall


# %%
# Load the viscous free fall example.
vff = viscous_free_fall.ViscousFreeFall()
distribution = vff.distribution
model = vff.model

# %%
# Restrict the number of inputs
# -----------------------------

# %%
# We define a function which has input only `z0` as input to *freeze* the 3 other inputs of the original model:
fz0 = ot.SymbolicFunction(["z0"], ["z0", "55", "80", "15"])

# %%
# Then we use the :class:`~openturns.PointToFieldConnection` to compose it with the original model.
model_z0 = ot.PointToFieldConnection(model, fz0)

# %%
# Sample trajectories
# --------------------

# %%
# In order to sample trajectories, we use the `getSample` method of the distribution of `z0` and apply the field function.
size = 10
inputSample = vff.distZ0.getSample(size)
outputSample = model_z0(inputSample)

# %%
# Draw viscous free fall trajectories.
graph = outputSample.drawMarginal(0)
graph.setTitle("Viscous free fall: %d trajectories" % (size))
graph.setXTitle(r"$t$")
graph.setYTitle(r"$z$")
view = otv.View(graph)
otv.View.ShowAll()



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()

# We create an empty Sample
sample = ot.Sample(0, 2)
sample.setName("EmptySample")
print("sample=", repr(sample))

try:
    # We access the element of the sample
    p = sample[0]
    print("p=", repr(p))

    # We should NEVER go here
    raise
except Exception:
    pass

# We create an small Sample
sample = ot.Sample(1, 2)
sample.setName("SmallSample")
print("sample=", repr(sample))
print(sample)
print(sample._repr_html_())

# We access the element of the sample
p = ot.Point(sample[0])
print("p=", repr(p))
try:
    # We try to access past the last element of the point
    p0, p1, p2 = p[0], p[1], p[2]
    print("p[0]=", p0, " p[1]=", p1, " p[2]=", p2)
    # We should NEVER go here
    raise
except Exception:
    pass

# We create a big Sample
sample = ot.Sample(1000000, 2)
sample.setName("BigSample")

# We populate the sample
for i in range(sample.getSize()):
    sample[i, 0] = i
    sample[i, 1] = i

print("sample first point=", repr(sample[0]))
print("sample last  point=", repr(sample[sample.getSize() - 1]))

# Test pretty-print or a sample with relatively large dimension
size = 49
dimension = 50
sample = ot.Sample(size, dimension)
sample.setName("BigSample50x50")
for i in range(size):
    sample[i] = ot.Point(list(range(dimension))) / size

print(sample._repr_html_())
